"""Almost a copy-paste from modules.training.core but adapted for reconstruction only."""

import os
import pickle

import torch
from torch import Tensor
from torch.utils.data import DataLoader
from tqdm import tqdm

from .core import Data, Model


def step(x: Tensor, model: Model, training: bool) -> float:
    x = x.to(model.device)
    x_hat = model.decoder(model.network.extract_features(x))
    criterion = model.train_criterion if training else model.val_criterion
    loss = criterion(x_hat, x)
    if training:
        model.optimizer.zero_grad()
        loss.backward()
        model.optimizer.step()
    return loss.detach().item()


def half_epoch(dataloader: DataLoader, model: Model, training=True) -> Model:
    model.network.train() if training else model.network.eval()
    model.decoder.train() if training else model.decoder.eval()
    torch.set_grad_enabled(training)
    loss, length = 0, 0
    for inputs, _ in dataloader:
        loss += step(inputs, model, training)
        length += len(inputs)
    loss /= length
    return model, loss


def full_epoch(data: Data, model: Model) -> Model:
    model, train_loss = half_epoch(data.train_loader, model, training=True)
    model, val_loss = half_epoch(data.val_loader, model, training=False)
    if model.scheduler is not None:
        model.scheduler.step()
    return model, train_loss, val_loss


def unsupervised_pretraining(
    data: Data, model: Model, num_epochs: int, output_dir: str
) -> Model:
    train_losses, val_losses = [], []
    with tqdm(total=num_epochs) as pbar:
        for i in range(num_epochs):
            pbar.set_description(f"EPOCH [{i+1}/{num_epochs}]")
            model, train_loss, val_loss = full_epoch(data, model)
            train_losses.append(train_loss)
            val_losses.append(val_loss)
            pbar.set_postfix({"train loss": train_loss, "val_loss": val_loss})
            pbar.update()
    with open(
        os.path.join(output_dir, "unsupervised_pretraining_train_losses.pickle"), "wb"
    ) as f:
        pickle.dump(train_losses, f)
    with open(
        os.path.join(output_dir, "unsupervised_pretraining_val_losses.pickle"), "wb"
    ) as f:
        pickle.dump(val_losses, f)
    return model



import numpy as np

from spfluo.utils.memory import split_batch


def test_split_batch():
    shapes = [(10,), (100,), (100, 120, 7), (100, 120, 7), (100, 120, 7), (1000,)]
    max_batch_sizes = [1, None, 10, 21 * 12 * 6, 12 * 7, 2048]

    tables = [np.zeros(s, dtype=int) for s in shapes]

    for t, max_batch in zip(tables, max_batch_sizes):
        for idx in split_batch(t.shape, max_batch):
            if type(idx) is tuple:
                idx = [idx]
            slices = [slice(i, j) for i, j in idx]
            t[tuple(slices)] += 1
            assert all(
                [
                    (j - i) <= max_batch if max_batch is not None else True
                    for (i, j) in idx
                ]
            )
        assert (t == 1).all()



from dataclasses import dataclass, field
from typing import Tuple

# +------------------------------------------------------------------------------------------+ #
# |                                           I/O                                            | #
# +------------------------------------------------------------------------------------------+ #


@dataclass
class IO:
    """Input/Output params.

    Args:

        point_cloud_path (str): Path to the point cloud csv. This point cloud will act as a
                                template to generate augmented particles.

        output_dir (str): Generated images and masks will be save in output_dir. One npz file per
                          pair will be created with two keys, 'image, and 'mask'.

        generated_dataset_size (int): How many pairs (image, mask) to generate.
    """

    point_cloud_path: str = "/data1/stage_Luc/inputs/sample_centriole_point_cloud.csv"
    output_dir: str = "/data1/stage_Luc/tilt/"
    generated_dataset_size: int = 25


# +------------------------------------------------------------------------------------------+ #
# |                                      Voxelisation                                        | #
# +------------------------------------------------------------------------------------------+ #


@dataclass
class Voxelisation:
    """Params controling the transformation from pointcloud to voxel grid.

    Args:

        num_particles (int): How many augmented copies to generate into each pair (image, mask).

        max_particle_dim (int): After voxelisation, each generated particle will be within an
                                irregular cube for which the side is at most max_particle_dim.

        bandwidth (int): The augmented point cloud will be converted into a voxel grid (ie a 3d
                         image) by a fast gaussian transorm algorithm. This parameters controls
                         the gaussian kernel size.

        epsilon (float): The desired error of the fast gaussian transform based voxelisation.

        cluster_range_xy (Tuple[int]): After voxelisation, particles will be placed onto the big
                                       3d image in a random center. This center will be sampled
                                       from a mixture of gaussian. Each gaussian standard
                                       deviation in x and y direction will be within
                                       radius_range_xy.

        cluster_range_z (Tuple[int]): After voxelisation, particles will be placed onto the big
                                      3d image in a random center. This center will be sampled
                                      from a mixture of gaussian. Each gaussian standard
                                      deviation in z direction will be within radius_range_z.

        nb_particles_per_cluster_range (Tuple[int]): Each particles cluster (ie each gaussian
                                                     used to sample center coordinates) will have
                                                     a number of particles inside within
                                                     nb_particles_per_cluster_range.
        tilt_strategy (str): Defines the distributions mixture used to sample tilt angles. Alike
                             {law}_{special_case} where law can be one of 'gaussian' or 'uniform'
                             and special_case can be top_side_only.
    """

    num_particles: int = 150
    max_particle_dim: int = 25
    bandwidth: int = 10
    epsilon: int = 1e-3
    cluster_range_xy: Tuple[int] = (100, 480)
    cluster_range_z: Tuple[int] = (5, 35)
    nb_particles_per_cluster_range: Tuple[int] = (45, 55)
    tilt_strategy: str = "uniform"


# +------------------------------------------------------------------------------------------+ #
# |                                      Augmentation                                        | #
# +------------------------------------------------------------------------------------------+ #


@dataclass
class Augmentation:
    """Params to control each generated particle augmentation.

    Args:

        shrink_range (Tuple[int]): The template point cloud will be shrinked by a random factor
                                       between shrink range.

        intensity_nb_holes_min (int): Non uniform density will be simulated by randomly deleting
                                      some points within spheres. This parameters controls the
                                      minimum amount of such sphere for each generated particle.

        intensity_nb_holes_max (int): Non uniform density will be simulated by randomly deleting
                                      some points within spheres. This parameters controls the
                                      maximum amount of such sphere for each generated particle.

        intensity_mean_ratio (int): Non uniform density will be simulated by randomly deleting
                                    some points within spheres of radius R. The random deletion
                                    follows a gaussian law: max proba in the center of the sphere,
                                    min proba in the edge. The mean of this gaussian law is
                                    R * mean_ratio.

        intensity_std_ratio_range (int): Non uniform density will be simulated by randomly
                                         deleting some points within spheres of radius R.
                                         The random deletion follows a gaussian law: max proba in
                                         the center of the sphere, min proba in the edge.
                                         The standard deviation of this gaussian law is:
                                         R * std_ratio.
                                         The std_ratio will be sampled uniformely within the
                                         specified range.

        rotation_proba (int): Each generated particle will be randomly rotated in 3d with a
                              probability rotation_proba. If the rotation occurs, the angles in
                              the 3 directions follow a uniform distribution from -180 to 180.
    """

    shrink_range: Tuple[int] = (0.8, 1.2)
    intensity_nb_holes_min: int = 1
    intensity_nb_holes_max: int = 8
    intensity_mean_ratio: float = 0.3
    intensity_std_ratio_range: float = (0.15, 0.25)
    rotation_proba: float = 1.0


# +------------------------------------------------------------------------------------------+ #
# |                                        Outliers                                          | #
# +------------------------------------------------------------------------------------------+ #


@dataclass
class Outliers:
    """Params controling outliers generation.

    Args:

        radius_range (Tuple[int]): Each outliers cluster will have a random radius within
                                   radius_range.
                                   Two radius are defined, one for the z-axis and one for the x
                                   and y axis because of the strong anistropy of resolution.

        nb_points_range (Tuple[int]): Each cluster will contain a number of points within
                                      nb_points_range.

        nb_clusters_range (Tuple[int]): The entire image will have a number of outliers clusters
                                        within nb_clusters_range.

        intensity_range (Tuple[int]): Each outlier point inside a cluster will have an intensity
                                      in intensity_range.
    """

    radius_range_xy: Tuple[int] = (25, 200)
    radius_range_z: Tuple[int] = (20, 40)
    nb_points_range: Tuple[int] = (200, 800)
    nb_clusters_range: Tuple[int] = (100, 200)
    intensity_range: Tuple[int] = (0.5, 1)


# +------------------------------------------------------------------------------------------+ #
# |                                         Sensor                                           | #
# +------------------------------------------------------------------------------------------+ #


@dataclass
class Sensor:
    """Params to control sensor faults: blur and noises.

    Args:

        anisotropic_blur (bool): Whether or not to add anisotropic blur to image and mask. If
                                 True, at the end of the generation process, the whole image and
                                 the whole mask will be convoluted with an anisotropic gaussian
                                 kernel.

        anisotropic_blur_sigma (Tuple[int]): This parameter controls the standard deviation of
                                             the gaussian blur for each axis, ordered as
                                             (z, y, x). E.g: (10*n, n, n) will lead to a strong
                                             z-axis resolution anisotropy.

        anisotropic_blur_border_mode (str): Because the anistropy of resolution is simulated by
                                            a kernel based convolution, the resulting image may
                                            be smaller than the original one. Thus the input
                                            image will be padded to prevent that. This parameter
                                            controls the padding strategy.
                                            Can be one of 'reflect', 'constant', 'nearest',
                                            'mirror', 'wrap'.

        gaussian_noise (bool): Whether to add gaussian noise to the final image or not.

        gaussian_noise_mean (float): If gaussian_noise is True, what will be the gaussian
                                     distribution's mean. As the generated image are float64 in
                                     range [0, 1], it is 0.5 by default.

        gaussian_noise_target_snr_db (int): What will be the Signal to Noise Ratio after the
                                            gaussian noise addition.

        poisson_noise (bool): Whether or not to add Poisson noise to the image. If both gaussian
                              and poisson are True, Poisson noise will be added after the
                              gaussian noise.
    """

    anisotropic_blur: bool = True
    anisotropic_blur_sigma: Tuple[int] = (5, 1, 1)
    anisotropic_blur_border_mode: str = "constant"
    gaussian_noise: bool = True
    gaussian_noise_mean: float = 0.5
    gaussian_noise_target_snr_db: int = 15
    poisson_noise: bool = True


# +------------------------------------------------------------------------------------------+ #
# |                                         Wrapper                                          | #
# +------------------------------------------------------------------------------------------+ #


@dataclass
class DataGenerationConfig:
    """This wraps all the parameters to generate pseudo random data defined abose.

    This config class will be used to instanciate a DataGenerator class which will use
    functional.py to perform its generation task.
    The basic idea is to generate many randomly augmented point cloud copies of a given particle
    point cloud, then convert them to images, then place them randomly into the output 3D image.

    Args:

        disable_cuda (bool): Usually, the device (cpu or gpu) will be automatically detected.
                            However it is possible to force cpu usage by setting this param to
                            True (for instance for debug purpose or for old GPU for which PyTorch
                            might not work as intended).

        target_shape (Tuple[int]): The shape of the final pair (image, mask) containing the
                                   generated copies of particles.

    Note that outliers is a Tuple of instance of the Outliers config class. This allows one to
    define several instances of Outliers config class with varying parameters. All those Outliers
    config will be used one after one in the DataGenerator class, so that several types of
    outliers can be generated.
    For instance, one may want to perform a first pass to add dense and small outliers clusters
    to  mimic artifacts, and a second pass with less dense but bigger outliers clusters to mimic
    light halo.
    """

    disable_cuda: bool = True
    target_shape: Tuple[int] = (128, 1024, 1024)
    io: IO = field(default_factory=IO)
    voxelisation: Voxelisation = field(default_factory=Voxelisation)
    augmentation: Augmentation = field(default_factory=Augmentation)
    outliers: Tuple[Outliers] = field(default_factory=tuple)
    sensor: Sensor = field(default_factory=Sensor)



from typing import Sequence

import torch
from torch import Tensor
from torch.distributions import Binomial


class BinomialGE(torch.nn.Module):
    def __init__(
        self, pi: float = None, slack: float = 1.0, entropy_penalty: float = 0.0
    ) -> None:
        super(BinomialGE, self).__init__()
        self.pi = pi
        self.slack = slack
        self.entropy_penalty = entropy_penalty
        self.criterion = torch.nn.BCEWithLogitsLoss()

    def forward(self, outputs: Tensor, targets: Tensor) -> Tensor:
        select = targets.data == 1
        classifier_loss = self.criterion(outputs[select], targets[select])
        # 1. Calculate Normal approximation to the distribution over positive count given by
        # the classifier
        select = targets.data == 0
        N = select.sum().item()
        p_hat = outputs[select].sigmoid()
        q_mu = p_hat.sum()
        q_var = torch.sum(p_hat * (1 - p_hat))
        count_vector = torch.arange(
            0, N + 1, device=targets.device, dtype=torch.float32
        )
        # below, 1e-10 is only a small epsilon to prevent NaN
        q_discrete = (-0.5 * (q_mu - count_vector) ** 2 / (q_var + 1e-10)).softmax(
            dim=0
        )
        # 2. KL of w from the binomial distribution with pi
        log_binom = Binomial(total_count=N, probs=self.pi).log_prob(count_vector)
        ge_penalty = -torch.sum(log_binom * q_discrete)
        if self.entropy_penalty > 0:
            q_entropy = 0.5 * (torch.log(q_var) + torch.log(2 * self.pi) + 1)
            ge_penalty += q_entropy * self.entropy_penalty
        loss = classifier_loss + self.slack * ge_penalty
        return loss


class MultiTaskLoss(torch.nn.Module):
    def __init__(
        self,
        criteria: Sequence[torch.nn.Module],
        eta: Sequence[float] = [0],
        learn_eta: bool = False,
        device: torch.device = torch.device("cpu"),
    ) -> None:
        """A module to compute a loss consisting of several terms, whose weights can be learnt.

        Args:
            criteria (Tuple[torch.nn.Module]): Tuple of losses, i.e. each terms constituting the
                                               total loss.
            eta (List[float], optional): Weighting of each losses. This MUST be equal in length
                                         to criteria. Defaults to [0].
            learn_eta (bool, optional): If False, weightings will remain constants during
                                        training. Otherwhise, they'll be optimized.
                                        Defaults to False.
            device (torch.device, optional): A torch device, typically cpu or gpu.
                                             Defaults to torch.device('cpu').
        """
        super(MultiTaskLoss, self).__init__()
        self.criteria = criteria
        self.eta = torch.Tensor(eta).to(device)
        if learn_eta:
            self.eta = torch.nn.Parameter(self.eta)

    def forward(
        self, outputs: Sequence[Tensor], targets: Sequence[Tensor]
    ) -> Sequence[Tensor]:
        loss = [
            criterion(o, y) for criterion, o, y in zip(self.criteria, outputs, targets)
        ]
        total_loss = torch.stack(loss) * torch.exp(-self.eta) + self.eta
        return loss, total_loss.sum()



import argparse
from pathlib import Path

from spfluo.ab_initio_reconstruction.api import AbInitioReconstruction
from spfluo.utils.log import base_parser, set_logging_level
from spfluo.utils.read_save_files import (
    read_image,
    read_images_in_folder,
    save_image,
)
from spfluo.utils.volume import interpolate_to_size, move_center_of_mass_to_center


def create_parser():
    parser = argparse.ArgumentParser("Ab initio reconstruction", parents=[base_parser])

    # Input files
    parser.add_argument("--particles_dir", type=str)
    parser.add_argument("--psf_path", type=str)

    # Ab initio params
    parser.add_argument("--M_axes", type=int, default=360**2)
    parser.add_argument("--M_rot", type=int, default=360)
    parser.add_argument("--dec_prop", type=float, default=1.2)
    parser.add_argument("--init_unif_prop", nargs=2, type=float, default=(1, 1))
    parser.add_argument("--coeff_kernel_axes", type=float, default=50.0)
    parser.add_argument("--coeff_kernel_rot", type=float, default=5.0)
    parser.add_argument("--eps", type=float, default=0)
    parser.add_argument("--lr", type=float, default=0.1)
    parser.add_argument("--N_axes", type=int, default=25)
    parser.add_argument("--N_rot", type=int, default=20)
    parser.add_argument("--prop_min", type=float, default=0)
    parser.add_argument("--interp_order", type=int, default=3)
    parser.add_argument("--N_iter_max", type=int, default=20)
    parser.add_argument("--gaussian_kernel", type=bool, default=True)
    parser.add_argument("--N_iter_with_unif_distr", type=int, default=None)
    parser.add_argument("--epochs_of_suppression", type=int, default=None)
    parser.add_argument("--proportion_of_views_suppressed", type=float, default=None)
    parser.add_argument("--convention", type=str, default="XZX")
    parser.add_argument("--dtype", type=str, default="float32")
    parser.add_argument("--reg_coeff", type=float, default=0)
    parser.add_argument("--beta_sampling", type=float, default=0)
    parser.add_argument("--epoch_length", type=int, default=None)
    parser.add_argument("--batch_size", type=int, default=1)
    parser.add_argument("--beta_grad", type=float, default=0)
    parser.add_argument("--random_sampling", type=bool, default=False)

    # GPU acceleration
    parser.add_argument("--gpu", action="store_true")
    parser.add_argument("--minibatch_size", type=int, default=None)

    # Ouput directory
    parser.add_argument("--output_dir", type=str)

    return parser


def main(args):
    # Read inputs
    # particles, names, images_metadata = read_images_in_folder(
    #     args.particles_dir, gpu=args.gpu, dtype=args.dtype
    # )
    # psf, psf_metadata = read_image(args.psf_path, gpu=args.gpu, dtype=args.dtype)
    particles, names = read_images_in_folder(
        args.particles_dir, gpu=args.gpu, dtype=args.dtype
    )
    psf = read_image(args.psf_path, gpu=args.gpu, dtype=args.dtype)

    # ensure_images_psf_isotropic_same_pixel_size(images_metadata, psf_metadata)

    # base_metadata = psf_metadata
    # assert particles.shape[1] == psf.shape[0] == 1, "input data should be monochannel"

    # squeeze data
    # particles = particles[:, 0]
    # psf = psf[0]

    ab_initio_params = dict(vars(args))
    for k in [
        "output_dir",
        "gpu",
        "psf_path",
        "particles_dir",
        "minibatch_size",
        "debug",
    ]:
        ab_initio_params.pop(k)
    reconstruction = AbInitioReconstruction(**ab_initio_params)
    psf = move_center_of_mass_to_center(interpolate_to_size(psf, particles.shape[1:]))
    reconstruction.fit(
        particles,
        psf=psf,
        output_dir=args.output_dir,
        minibatch_size=args.minibatch_size,
        particles_names=names,
    )
    save_image(
        str(Path(args.output_dir) / "final_reconstruction.ome.tiff"),
        reconstruction._volume,
        order="ZYX",
        #    metadata=base_metadata,
    )


if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()
    set_logging_level(args)
    main(args)



from __future__ import annotations

import json
import os
import shutil
from typing import TYPE_CHECKING, Any, Callable, Literal, Tuple

import numpy as np
import tifffile
from numpy.core.numeric import normalize_axis_tuple
from skimage import io
from skimage.metrics import structural_similarity as ssim
from tqdm.auto import tqdm

from spfluo.ab_initio_reconstruction.volume_representation.pixel_representation import (
    Fourier_pixel_representation,
)
from spfluo.utils.array import array_namespace, get_device, numpy, to_numpy
from spfluo.utils.loading import read_poses, save_poses
from spfluo.utils.memory import split_batch
from spfluo.utils.transform import get_transform_matrix
from spfluo.utils.volume import fourier_shift, phase_cross_correlation

from ...utils.read_save_files import make_dir, save_image
from ..common_image_processing_methods.others import normalize, stopping_criteria
from ..common_image_processing_methods.rotation_translation import (
    conversion_2_first_eulers_angles_cartesian,
    rotation,
)
from ..volume_representation.gaussian_mixture_representation.GMM_grid_evaluation import (  # noqa: E501
    one_d_gaussian,
)

if TYPE_CHECKING:
    from spfluo.utils.array import Array

if TYPE_CHECKING:
    from spfluo.utils.array import Device, array_api_module


def gd_importance_sampling_3d(
    volume_representation: Fourier_pixel_representation,
    uniform_sphere_discretization,
    true_trans_vecs,
    views,
    imp_distrs_axes,
    imp_distrs_rot,
    unif_prop,
    unif_prop_min,
    params_learning_alg,
    output_dir,
    ground_truth=None,
    folder_views_selected=None,
    xp: "array_api_module" = numpy,
    device: "Device" = "cpu",
    minibatch_size=None,
    callback: Callable[[np.ndarray, int], Any] | None = None,
    particles_names: list[str] | None = None,
    keep_best: bool = True,  # if False, keep last iteration
):
    params_to_save = params_learning_alg.__dict__.copy()
    del params_to_save["params"]
    del params_to_save["dtype"]
    make_dir(output_dir)
    with open(os.path.join(output_dir, "params_learning_alg.json"), "w") as f:
        json.dump(params_to_save, f)
    imp_distrs_axes, imp_distrs_rot = to_numpy(imp_distrs_axes, imp_distrs_rot)
    thetas, phis, psis = to_numpy(*uniform_sphere_discretization[0])
    views = to_numpy(views)
    unif_prop_axes, unif_prop_rot = unif_prop
    epoch_length = (
        params_learning_alg.epoch_length
        if params_learning_alg.epoch_length is not None
        else len(views)
    )
    bs = params_learning_alg.batch_size

    if folder_views_selected is None:
        folder_views_selected = f"{output_dir}/views_selected"
        make_dir(folder_views_selected)

    if particles_names is None:
        particles_names = [str(i) for i in range(len(views))]

    x, y, z = conversion_2_first_eulers_angles_cartesian(thetas, phis)
    axes = np.array([x, y, z])
    M_axes = len(thetas)
    M_rot = len(psis)
    recorded_energies = []
    energies_each_view = [[] for _ in range(len(views))]
    itr = 0

    recorded_shifts = [[] for _ in range(len(views))]
    sub_dir = os.path.join(output_dir, "intermediar_results")
    make_dir(sub_dir)
    ests_poses = []
    nb_step_of_supress = 0
    pbar = tqdm(total=params_learning_alg.N_iter_max, leave=False, desc="energy : +inf")
    views = views.astype(params_learning_alg.dtype)
    N_axes, N_rot = params_learning_alg.N_axes, params_learning_alg.N_rot
    minimum_energy, best_itr = None, None
    while itr < params_learning_alg.N_iter_max and (
        not stopping_criteria(recorded_energies, params_learning_alg.eps)
    ):
        # print(f'nb views epoch {itr} : ', len(views))

        nb_views = len(views)
        itr += 1
        total_energy = 0
        estimated_poses_iter = np.zeros((nb_views, 6))

        if params_learning_alg.random_sampling:
            # Weighted-random sampling
            weights = np.ones((nb_views,)) / nb_views
            last_energies = np.array(
                [
                    energies_each_view[v][-1]
                    if len(energies_each_view[v]) > 0
                    else np.inf
                    for v in range(len(views))
                ]
            )
            m = last_energies.max()
            if np.isinf(m):
                weights = np.array([1.0 if np.isinf(e) else 0.0 for e in last_energies])
                weights /= weights.sum()
            else:
                last_energies_centered = last_energies - m
                weights = (
                    np.exp(params_learning_alg.beta_sampling * last_energies_centered)
                    / np.exp(
                        params_learning_alg.beta_sampling * last_energies_centered
                    ).sum()
                )  # softmax
            chosen_views = np.random.choice(nb_views, size=epoch_length, p=weights)
        else:
            chosen_views = np.arange(epoch_length) % len(views)
            np.random.shuffle(chosen_views)
        batches = [
            chosen_views[i * bs : (i + 1) * bs] for i in range(epoch_length // bs)
        ]
        last_batch_size = epoch_length % bs
        if last_batch_size > 0:
            batches += [chosen_views[-last_batch_size:]]

        pbar2 = tqdm(batches, leave=False)
        for batch in pbar2:
            gradients_batch = []
            energies_batch = []
            for v in batch:
                view = views[v]
                # Pick a subset of the discretization of SO(3) based on
                # importance distributions
                indices_axes = np.random.choice(
                    range(M_axes), p=imp_distrs_axes[v], size=N_axes
                )
                indices_rot = np.random.choice(
                    range(M_rot), p=imp_distrs_rot[v], size=N_rot
                )
                rot_vecs = np.stack(
                    np.broadcast_arrays(
                        thetas[indices_axes][:, None],
                        phis[indices_axes][:, None],
                        psis[indices_rot],
                    ),
                    axis=-1,
                )  # the euler angles, shape N_axes x N_rot x 3
                image_shape = view.shape
                # turn the euler angles into transform matrices, shape N_axes*N_rot x 3
                transforms = get_transform_matrix(
                    shape=image_shape,
                    euler_angles=rot_vecs.reshape(-1, 3),
                    translation=np.zeros((N_axes * N_rot, 3)),
                    convention=params_learning_alg.convention,
                    degrees=True,
                )
                energies = np.empty((transforms.shape[0],))

                view = xp.asarray(
                    view, device=device, dtype=getattr(xp, params_learning_alg.dtype)
                )
                # transforms represents transformation from reference volume
                # to the views
                transforms = xp.asarray(
                    transforms,
                    device=device,
                    dtype=getattr(xp, params_learning_alg.dtype),
                )
                reference = xp.asarray(
                    xp.fft.ifftn(volume_representation.volume_fourier).real,
                    device=device,
                )
                reference_fft = volume_representation.volume_fourier
                psf = xp.asarray(volume_representation.psf, device=device)
                for start, end in split_batch(
                    (N_axes * N_rot,), max_batch=minibatch_size
                ):
                    transforms_minibatch = transforms[start:end]

                    # Compute shifts
                    # views are transformed back to the reference volume
                    # then phase cross correlation is computed to get the shifts
                    view_ = view
                    (
                        shifts_minibatch,
                        (
                            psf_inverse_transformed_fft_minibatch,
                            view_inverse_transformed_fft_minibatch,
                        ),
                    ) = compute_shifts(
                        reference,
                        psf,
                        transforms_minibatch,
                        view_,
                        interp_order=1,
                    )

                    # Compute the energy associated with each transformation
                    energies_minibatch = compute_energy(
                        reference_fft,
                        psf_inverse_transformed_fft_minibatch,
                        view_inverse_transformed_fft_minibatch,
                        shift=shifts_minibatch,
                        space="fourier",
                    )
                    del view_inverse_transformed_fft_minibatch

                    # save results to cpu
                    energies[start:end] = to_numpy(energies_minibatch)

                    del shifts_minibatch, energies_minibatch

                # Some reshaping
                energies = energies.reshape(len(indices_axes), len(indices_rot))
                transforms = xp.reshape(
                    transforms, (len(indices_axes), len(indices_rot), 4, 4)
                )

                # Find the energy minimum
                j, k = np.unravel_index(np.argmin(energies), energies.shape)
                min_energy = energies[j, k]
                energies_batch.append(min_energy)  # store it
                energies_each_view[v].append(min_energy)
                pbar2.set_description(
                    f"particle {particles_names[v]} energy : {min_energy:.1f}"
                )

                # compute the shift of the minimum
                # get also the psf and the view rotated for the computation
                # of the gradient

                (
                    shift,
                    (psf_inverse_transformed_fft, view_inverse_transformed_fft),
                ) = compute_shifts(
                    xp.asarray(
                        xp.fft.ifftn(volume_representation.volume_fourier),
                        device=device,
                    ),
                    volume_representation.psf,
                    transforms[j, k][None],
                    view,
                    interp_order=1,
                )
                shift, psf_inverse_transformed_fft, view_inverse_transformed_fft = (
                    shift[0],
                    psf_inverse_transformed_fft[0],
                    view_inverse_transformed_fft[0],
                )
                view_inverse_transformed_shifted_fft = fourier_shift(
                    view_inverse_transformed_fft, shift
                )

                # Recover the pose of the view

                # To go from view to the reference volume, we did:
                # an inverse transform, which is a pure rotation lets call it R^-1
                # then a shift, T_{t}

                # So, volume = (T_{t} o R^-1)(view)
                # If we invert, (R o T_{-t})(volume) = view
                # Equivalently, (T_{- R^-1 x t} o R)(volume) = view

                # The associated pose with the view is therefore
                translation = (
                    -np.linalg.inv(to_numpy(transforms[j, k]))
                    @ np.concatenate((to_numpy(shift), [0.0]), axis=0)[:, None]
                )

                # The transformation associated with that minimum
                best_idx_axes, best_idx_rot = indices_axes[j], indices_rot[k]
                rot_vec = np.asarray(
                    [
                        thetas[best_idx_axes],
                        phis[best_idx_axes],
                        psis[best_idx_rot],
                    ]
                )
                estimated_poses_iter[v, :3] = rot_vec  # store the rotation
                estimated_poses_iter[v, 3:] = translation[
                    :3, 0
                ]  # store the translation

                # Compute the gradient of the energy with respect to the volume
                grad = compute_grad(
                    xp.asarray(volume_representation.volume_fourier, device=device),
                    psf_inverse_transformed_fft,
                    view_inverse_transformed_shifted_fft,
                )
                gradients_batch.append(to_numpy(grad))  # accumulate

                # Update the importance distributions
                energies = normalize(energies, max=6)
                likelihoods = np.exp(-energies)
                phi_axes = likelihoods.dot(1 / imp_distrs_rot[v][indices_rot]) / N_rot
                phi_rot = (
                    likelihoods.T.dot(1 / imp_distrs_axes[v][indices_axes]) / N_axes
                )
                K_axes = np.exp(
                    params_learning_alg.coeff_kernel_axes
                    * axes[:, indices_axes].T.dot(axes)
                )
                K_rot = np.zeros((N_rot, M_rot))
                for k, idx_rot in enumerate(indices_rot):
                    a = psis[idx_rot]
                    if params_learning_alg.gaussian_kernel:
                        K_rot[k, :] = one_d_gaussian(
                            psis, a, params_learning_alg.coeff_kernel_rot
                        )
                    else:
                        K_rot[k, :] = np.exp(
                            np.cos(a - psis) * params_learning_alg.coeff_kernel_rot
                        )
                update_imp_distr(
                    imp_distrs_axes, phi_axes, K_axes, unif_prop_axes, M_axes, v
                )
                update_imp_distr(
                    imp_distrs_rot, phi_rot, K_rot, unif_prop_rot, M_rot, v
                )

            # Weighted gradient descent
            energies_batch = np.array(energies_batch)
            m = energies_batch.max()
            if np.isinf(m):
                raise ValueError("Ab initio reconstruction diverged")
            e = np.exp(-params_learning_alg.beta_grad * (energies_batch - m))
            grad_weights = e / e.sum()
            grad = (grad_weights * np.stack(gradients_batch, axis=-1)).sum(axis=-1)
            volume_representation.gd_step(
                volume_representation.xp.asarray(grad),
                params_learning_alg.lr,
                params_learning_alg.reg_coeff,
            )

            # Center of mass centered
            shift = to_numpy(volume_representation.center())
            for v in range(estimated_poses_iter.shape[0]):
                estimated_poses_iter[v, 3:] -= to_numpy(shift)

            # Increase energy
            total_energy += np.sum(energies_batch)

            if callback:
                callback(
                    volume_representation.get_image_from_fourier_representation(), itr
                )

        ests_poses.append(estimated_poses_iter)
        pbar2.close()

        if (
            params_learning_alg.epochs_of_suppression is not None
            and len(params_learning_alg.epochs_of_suppression) > 0
            and itr == params_learning_alg.epochs_of_suppression[0]
        ):
            nb_step_of_supress += 1
            prop_to_suppress = params_learning_alg.proportion_of_views_suppressed.pop(0)
            nb_views_to_suppress = int(len(views) * prop_to_suppress)
            params_learning_alg.epochs_of_suppression.pop(0)
            energies_each_views_current_iter = np.array(energies_each_view)[:, -1]
            # print('energies each views', energies_each_views_current_iter)
            idx_views_to_keep = np.argsort(energies_each_views_current_iter)[
                : len(energies_each_views_current_iter) - nb_views_to_suppress
            ]
            # print('idx kepts', idx_views_to_keep)
            views = [views[idx] for idx in idx_views_to_keep]
            imp_distrs_axes = [imp_distrs_axes[idx] for idx in idx_views_to_keep]
            imp_distrs_rot = [imp_distrs_rot[idx] for idx in idx_views_to_keep]
            energies_each_view = [energies_each_view[idx] for idx in idx_views_to_keep]
            recorded_shifts = [recorded_shifts[idx] for idx in idx_views_to_keep]
            particles_names = [particles_names[idx] for idx in idx_views_to_keep]
            folder_views_selected_step = (
                f"{folder_views_selected}/step_{nb_step_of_supress}"
            )
            make_dir(folder_views_selected_step)
            for i, fn in enumerate(particles_names):
                save_image(f"{folder_views_selected_step}/{fn}", views[i])

        # Register reconstrution with groundtruth and save it
        volume_representation.register_and_save(
            sub_dir,
            f"recons_epoch_{itr}.ome.tiff",
            ground_truth=ground_truth,
        )

        # Update uniform distribution
        unif_prop_axes /= params_learning_alg.dec_prop
        unif_prop_rot /= params_learning_alg.dec_prop
        if params_learning_alg.N_iter_with_unif_distr is not None:
            if itr > params_learning_alg.N_iter_with_unif_distr:
                unif_prop_axes, unif_prop_rot = 0, 0
        if unif_prop_axes < unif_prop_min:
            unif_prop_axes = unif_prop_min
        if unif_prop_rot < unif_prop_min:
            unif_prop_rot = unif_prop_min

        # Save stuff
        save_poses(
            f"{sub_dir}/estimated_poses_epoch_{itr}.csv",
            estimated_poses_iter,
            particles_names,
        )
        if ground_truth is not None:
            regist_im = io.imread(os.path.join(sub_dir, f"recons_epoch_{itr}.tif"))
            ssim_gt_recons = ssim(normalize(ground_truth), normalize(regist_im))
            with open(os.path.join(output_dir, "ssims.csv"), "a") as f:
                f.write(f"{ssim_gt_recons}\n")

        if not os.path.exists(
            distributions_angles_dir := os.path.join(output_dir, "distributions_angles")
        ):
            os.makedirs(distributions_angles_dir)

        np.save(
            os.path.join(distributions_angles_dir, f"iter={itr:04}_rot.npy"),
            imp_distrs_rot,
        )
        np.save(
            os.path.join(distributions_angles_dir, f"iter={itr:04}_axes.npy"),
            imp_distrs_axes,
        )

        if not os.path.exists(energies_dir := os.path.join(output_dir, "energies")):
            os.makedirs(energies_dir)
        np.save(
            os.path.join(energies_dir, f"energies_each_view_iter={itr:04}.npy"),
            np.array(energies_each_view),
        )

        total_energy /= epoch_length
        with open(os.path.join(output_dir, "energies.csv"), "a") as f:
            f.write(f"{total_energy}\n")
        recorded_energies.append(total_energy)

        if minimum_energy is None or total_energy < minimum_energy:
            minimum_energy = total_energy
            best_itr = itr

        pbar.set_description(f"energy : {total_energy:.1f}")
        pbar.update()

    if itr > 0:
        itr = best_itr if keep_best else itr
        shutil.copyfile(
            os.path.join(sub_dir, f"recons_epoch_{itr}.ome.tiff"),
            os.path.join(output_dir, "final_recons.ome.tiff"),
        )
        shutil.copyfile(
            os.path.join(sub_dir, f"estimated_poses_epoch_{itr}.csv"),
            os.path.join(output_dir, "poses.csv"),
        )
    pbar.close()

    best_image = tifffile.imread(os.path.join(sub_dir, f"recons_epoch_{itr}.ome.tiff"))
    best_energies_each_view = np.load(
        os.path.join(output_dir, "energies", f"energies_each_view_iter={itr:04}.npy")
    )
    best_estimated_poses, _ = read_poses(
        os.path.join(sub_dir, f"estimated_poses_epoch_{itr}.csv")
    )

    return (
        recorded_energies,
        recorded_shifts,
        unif_prop,
        best_image,
        itr,
        best_energies_each_view,
        views,
        particles_names,
        best_estimated_poses,
    )


def update_imp_distr(imp_distr, phi, K, prop, M, v):
    # phi = phi ** (1 / temp)
    q_first_comp = phi @ K
    q_first_comp /= np.sum(q_first_comp)
    imp_distr[v] = (1 - prop) * q_first_comp + prop * np.ones(M) / M
    return q_first_comp


def compute_shifts(
    reference: "Array",
    psf: "Array",
    transforms: "Array",
    view: "Array",
    interp_order: int = 1,
) -> Tuple["Array", "Array", "Array"]:
    xp = array_namespace(transforms, view, psf, view)

    inverse_transforms = xp.linalg.inv(transforms)

    device = get_device(inverse_transforms)
    assert device == get_device(view)
    reference_fft = xp.fft.fftn(reference)

    N = inverse_transforms.shape[0]
    image_shape = psf.shape

    # Rotate the psf backward
    psfs_rotated = rotation(
        xp.broadcast_to(psf, (N,) + image_shape),
        inverse_transforms,
        order=interp_order,
    )
    psfs_rotated_fft = xp.fft.fftn(psfs_rotated, axes=(1, 2, 3))

    # Rotate the view back to the volume
    view_rotated = rotation(
        xp.broadcast_to(view, (N,) + image_shape),
        inverse_transforms,
        order=interp_order,
    )
    view_rotated_fft = xp.fft.fftn(view_rotated, axes=(1, 2, 3))

    shift, _, _ = phase_cross_correlation(
        psfs_rotated_fft * reference_fft,
        view_rotated_fft,
        nb_spatial_dims=3,
        upsample_factor=10,
        normalization=None,
        space="fourier",
    )
    shift = xp.stack(shift, axis=-1)
    return shift, (psfs_rotated_fft, view_rotated_fft)


def compute_energy(
    reference: "Array",
    psf: "Array",
    view: "Array",
    transform: "Array | None" = None,
    shift: "Array | None" = None,
    space: Literal["fourier", "real"] = "fourier",
    order: int = 1,
):
    """Compute the energy for a given reference, psf and view
    If the transform is not given, the psf and the view are considered to be rotated.
    If the shift is not given, the view is considered to be shifted.
    """
    xp = array_namespace(view)
    device = get_device(view)
    image_shape = psf.shape[-3:]
    if transform is not None:
        N = transform.shape[0]
        inverse_transform = xp.linalg.inv(transform)
        view_real = view if space == "real" else xp.fft.ifftn(view).real
        view_rotated_real = rotation(
            xp.broadcast_to(view_real, (N,) + image_shape),
            inverse_transform,
            order=order,
        )
        view_rotated_fft = xp.fft.fftn(view_rotated_real, axes=(1, 2, 3))

        psf_real = psf if space == "real" else xp.fft.ifftn(psf).real
        psf_rotated_real = rotation(
            xp.broadcast_to(psf_real, (N,) + image_shape),
            inverse_transform,
            order=order,
        )
        psf_rotated_fft = xp.fft.fftn(psf_rotated_real, axes=(1, 2, 3))
    else:
        view_rotated_fft = view if space == "fourier" else xp.fft.fftn(view)
        psf_rotated_fft = psf if space == "fourier" else xp.fft.fftn(psf)

    if shift is not None:
        view_rotated_shifted_fft = fourier_shift(view_rotated_fft, shift)
    else:
        view_rotated_shifted_fft = view_rotated_fft

    if space == "real":
        reference_fft = xp.fft.fftn(reference)
    else:
        reference_fft = reference

    reference_volume_fft = xp.asarray(reference_fft, device=device)
    psf_rotated_fft = xp.asarray(psf_rotated_fft, device=device)
    image_size = xp.prod(xp.asarray(image_shape, device=device))

    def vector_norm(x: "Array", axis: tuple[int]):
        normalized_axis = normalize_axis_tuple(axis, x.ndim)
        rest = tuple(i for i in range(x.ndim) if i not in normalized_axis)
        newshape = axis + rest
        x = xp.reshape(
            xp.permute_dims(x, newshape),
            (
                int(np.prod([x.shape[i] for i in axis], dtype=int)),
                *[x.shape[i] for i in rest],
            ),
        )
        return xp.linalg.vector_norm(x, axis=0, ord=2)

    energy = (
        vector_norm(
            psf_rotated_fft * reference_volume_fft - view_rotated_shifted_fft,
            axis=(-3, -2, -1),
        )
        ** 2
        / image_size
    )
    return energy


def compute_grad(volume_fourier, psf_rotated_fft, view_rotated_fft):
    grad = psf_rotated_fft * (psf_rotated_fft * volume_fourier - view_rotated_fft)
    return grad



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Optional

import numpy as np

from spfluo.ab_initio_reconstruction.common_image_processing_methods.others import (
    normalize,
)
from spfluo.utils.array import array_namespace, get_device
from spfluo.utils.volume import (
    discretize_sphere_uniformly,
)

from .learning_algorithms.gradient_descent_importance_sampling import (
    gd_importance_sampling_3d,
)
from .params import ParametersMainAlg
from .volume_representation.pixel_representation import Fourier_pixel_representation

if TYPE_CHECKING:
    from spfluo.utils.array import Array


class AbInitioReconstruction:
    def __init__(
        self, callback: Callable[[np.ndarray, int], Any] | None = None, **params
    ):
        self.params = params
        self._volume = None
        self._energies = None
        self._num_iter = None
        self._poses = None
        self.callback = callback

    def fit(
        self,
        X: "Array",
        psf: Optional["Array"] = None,
        output_dir: Optional[str] = None,
        minibatch_size: Optional[int] = None,
        particles_names: Optional[list[str]] = None,
    ):
        """Reconstruct a volume based on views of particles"""
        if psf is None:
            raise NotImplementedError  # TODO : default psf to gaussian
        if output_dir is None:
            output_dir = "./ab-initio-output"

        params_learning_alg = ParametersMainAlg(**self.params)
        fourier_volume = Fourier_pixel_representation(
            3,
            psf.shape[0],
            psf,
            init_vol=None,
            random_init=True,
            dtype=params_learning_alg.dtype,
        )

        N = X.shape[0]
        # normalize views
        xp = array_namespace(X)
        X = xp.stack([normalize(X[i]) for i in range(N)])

        uniform_sphere_discretization = discretize_sphere_uniformly(
            xp,
            params_learning_alg.M_axes,
            params_learning_alg.M_rot,
            dtype=xp.float64,
        )
        imp_distrs_axes = (
            xp.ones((N, params_learning_alg.M_axes)) / params_learning_alg.M_axes
        )
        imp_distrs_rot = (
            xp.ones((N, params_learning_alg.M_rot)) / params_learning_alg.M_rot
        )

        (
            recorded_energies,
            recorded_shifts,
            unif_prop,
            reconstruction,
            itr,
            energies_each_view,
            views,
            file_names,
            ests_poses,
        ) = gd_importance_sampling_3d(
            fourier_volume,
            uniform_sphere_discretization,
            None,
            X,
            imp_distrs_axes,
            imp_distrs_rot,
            params_learning_alg.init_unif_prop,
            0,
            params_learning_alg,
            output_dir,
            ground_truth=None,
            folder_views_selected=None,
            xp=xp,
            device=get_device(X),
            minibatch_size=minibatch_size,
            callback=self.callback,
            particles_names=particles_names,
        )
        self._volume = xp.asarray(reconstruction)
        self._energies = xp.mean(xp.asarray(energies_each_view), axis=0)
        self._num_iter = itr
        self._poses = xp.asarray(ests_poses)

        return self



import numpy as np


def gaussian_mixture_isotrop_identical_gaussians(
    grid, coeffs, centers, sigma, nb_dim, alpha
):
    covs = [sigma**2 * np.eye(nb_dim) for _ in range(len(coeffs))]
    gauss_mixture, _ = gaussian_mixture(grid, coeffs, centers, covs, nb_dim, alpha)
    return gauss_mixture


def gaussian_mixture(
    grid, coeffs, centers, covs, nb_dim, alpha, return_gaussians=False
):
    """computes the values of a multivariate gaussian mixture on a 3d grid.
    It truncates the gaussians around their centers
    IN :
    - grid : 4d array of shape  (d, h, w, 3) set of 3d voxels organised in a 3d volume.
             Values are computed on these voxels
    - centers : array of shape (Nb_gaussians, 3,1)
    - covs : covariance matrices of the gaussians, shape (Nb_gaussians, 3,3)
    - coeffs : linear coefficient of each gaussian
    OUT : array of shape (d, h, w) containing the values taken
          by the gaussian mixture at the voxels on the grid
    """
    shp = tuple([grid.shape[d] for d in range(nb_dim)])
    gauss_mixture = np.zeros(shp)
    gaussians = []
    for i in range(coeffs.shape[0]):
        gaussian = truncated_nd_gaussian(
            grid,
            gauss_mixture,
            coeffs[i],
            centers[i],
            covs[i],
            nb_dim,
            alpha,
            return_gaussians,
        )
        if return_gaussians:
            gaussians.append(gaussian)
    return gauss_mixture, gaussians


def truncated_nd_gaussian(
    grid, gaussian_mixture, coeff, center, cov, nb_dim, alpha, return_gaussian=False
):
    """computes the values of a multivariate gaussian on a 3d grid.
    It truncates the gaussian around the center
    IN :
    - grid : 4d array of shape (d, h, w, 3) set of 3d voxels organised in a 3d volume.
             Values are computed on these voxels
    - center : array of shape (3)
    - cov : covariance matrix of the gaussian, shape (3,3)
    - half_size_cut : half_size of the cut
    OUT : array of shape (d, h, w) containing the values taken by the gaussian at
          the voxels on the grid
    """
    grid_step = 2 / (len(grid) - 1)
    center_int = [int((center[d] - 1) / grid_step + len(grid)) for d in range(nb_dim)]
    # center_int = [int(grid_reverse_transform(center[d])) for d in range(nb_dim)]
    mins = []
    maxs = []
    lenghts = []
    slices = []
    for d in range(nb_dim):
        half_size_cut = int(alpha * np.sqrt(cov[d, d]) / grid_step + 1)
        mi = max(0, center_int[d] - half_size_cut)
        mi = min(mi, grid.shape[d])
        mins.append(mi)
        ma = min(grid.shape[d], center_int[d] + half_size_cut)
        ma = max(0, ma)
        maxs.append(ma)
        lenghts.append(ma - mi)
        slices.append(slice(mi, ma, 1))

    slices = tuple(slices)
    cut_grid = grid[slices]
    res_cut_grid = nd_gaussian(cut_grid, center, cov, nb_dim)
    shp = tuple([grid.shape[d] for d in range(nb_dim)])
    gaussian_mixture[slices] += coeff * res_cut_grid
    if return_gaussian:
        gaussian = np.zeros(shp)
        gaussian[slices] = res_cut_grid
        return gaussian
    return 0


def one_d_gaussian(points, center, sig):
    centered_grid = points - center
    A = centered_grid**2 / sig**2
    return np.exp(-A / 2)


def nd_gaussian_with_points(points, center, cov):
    centered_grid = points - center
    centered_grid = centered_grid.T
    inv_cov = np.linalg.inv(cov)
    center_grid_dot_cov = inv_cov @ centered_grid
    A = (center_grid_dot_cov * centered_grid).sum(axis=0)
    gauss = np.exp(-A / 2)
    return gauss


def flattens_grid(grid, nb_dim):
    shp = grid.shape
    size_flatten = 1
    for i in range(nb_dim):
        size_flatten *= shp[i]
    flatten_grid = grid.reshape((size_flatten, nb_dim))
    return flatten_grid, shp


def nd_gaussian(grid, center, cov, nb_dim):
    flatten_grid, shp = flattens_grid(grid, nb_dim)
    gauss = nd_gaussian_with_points(flatten_grid, center, cov)
    shape_unflatten = tuple([shp[i] for i in range(nb_dim)])
    return gauss.reshape(shape_unflatten)


def make_grid(size, nb_dim, mi=-1, ma=1):
    slices = []
    for d in range(nb_dim):
        slices.append(slice(0, size, 1))
    slices = tuple(slices)
    transpose_idx = list(range(1, nb_dim + 1))
    transpose_idx.append(0)
    grid = np.mgrid[slices].transpose(*transpose_idx)
    if mi is not None and ma is not None:
        grid_step = (ma - mi) / (size - 1)
        grid = grid_step * (grid - size + 1) + ma
    return grid



"""This file defines functions to add random augmentation to a 3d point cloud in order to
generate data. It expects torch.Tensor point clouds and should operate indifferently on cpu
and gpu.
"""

from typing import Tuple, Union

import numpy as np
import torch
from scipy.spatial.transform import Rotation
from scipy.stats import multivariate_normal

# _______________________________________________________________________________________________ #


def get_center_and_radius(pointcloud: torch.Tensor) -> Tuple[torch.Tensor, float]:
    """Returns the center and the radius of a given point cloud.
    The center is the mean coordinate.
    The radius is given by the max distance between the center and any point from the point cloud.

    Args:
        pointcloud (torch.Tensor): Point cloud tensor, of shape (N, 3).

    Returns:
        Tuple[torch.Tensor, float]: * The center coordinates (x, y, z), of shape (3, ).
                                    * The radius as a single float.
    """
    center = pointcloud.mean(dim=0)
    distances = torch.norm(pointcloud - center, dim=1)
    return center, distances.max().cpu().numpy().item()


# _______________________________________________________________________________________________ #


def get_FOV(pointcloud: Union[torch.Tensor, np.ndarray]) -> Tuple[float]:
    """Get the Field Of View of a point cloud.

    Args:
        pointcloud (Union[torch.Tensor, np.ndarray]): Point cloud array like, of shape (N, 3).

    Returns:
        Tuple[float]: x_min, x_max, y_min, y_max, z_min, z_max.
    """
    return (
        pointcloud[:, 0].min(),
        pointcloud[:, 0].max(),
        pointcloud[:, 1].min(),
        pointcloud[:, 1].max(),
        pointcloud[:, 2].min(),
        pointcloud[:, 2].max(),
    )


# _______________________________________________________________________________________________ #


def shrink(pointcloud: torch.Tensor, factor: float = 1500.0) -> torch.Tensor:
    """Shrink a point cloud tensor by a given factor and center it.

    Args:
        pointcloud (torch.Tensor): Point cloud tensor to be shrinked, of shape (N, 3). SHOULD BE CENTERED!
        factor (float, optional): Shrink factor. Defaults to 1500.0.

    Returns:
        torch.Tensor: Shrinked and centered point cloud tensor.
    """
    return pointcloud / factor


# _______________________________________________________________________________________________ #


def rotate(
    pointcloud: torch.Tensor,
    rotation_angles: Tuple[float] = None,
    device: torch.device = torch.device("cpu"),
) -> torch.Tensor:
    """Randomly rotate a pointcloud with a given probability. If no angles are provided, the
        rotation is determined by uniformly sampling 3 Euler's angles.

    Args:
        pointcloud (torch.Tensor): Point cloud tensor, of shape (N, 3).
        rotation_angles (Tuple[int]): zyx Euler's angles. If None, will be sampled uniformely in
                                      [0, 360]. Defaults to None.
        p (float, optional): Rotation probability. Defaults to 0.5.
        device (torch.device, optional): CPU or GPU. Defaults to torch.device('cpu').

    Returns:
        torch.Tensor: * Rotated point cloud tensor, of shape (N, 3).
    """
    if rotation_angles is None:
        rotation_angles = np.random.uniform(low=0, high=360, size=3)
    rot = Rotation.from_euler("XZX", rotation_angles, degrees=True)
    rot_mat = torch.as_tensor(rot.as_matrix(), dtype=pointcloud.dtype, device=device)
    return (rot_mat @ pointcloud.T).T


# _______________________________________________________________________________________________ #


def translate(
    pointcloud: torch.Tensor,
    translation: Tuple[float] = None,
    device: torch.device = torch.device("cpu"),
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Randomly rotate a pointcloud with a given probability. If no angles are provided, the
        rotation is determined by uniformly sampling 3 Euler's angles.

    Args:
        pointcloud (torch.Tensor): Point cloud tensor, of shape (N, 3).
        rotation_angles (Tuple[int]): zyx Euler's angles. If None, will be sampled uniformely in
                                      [0, 360]. Defaults to None.
        p (float, optional): Rotation probability. Defaults to 0.5.
        device (torch.device, optional): CPU or GPU. Defaults to torch.device('cpu').

    Returns:
        torch.Tensor: * Rotated point cloud tensor, of shape (N, 3).
    """
    if translation is None:
        return pointcloud
    t = torch.as_tensor(translation, dtype=pointcloud.dtype, device=pointcloud.device)
    return pointcloud + t


# _______________________________________________________________________________________________ #


def uniform_in_sphere(
    center: torch.Tensor,
    radius: float,
    N: int = 1,
    device: torch.device = torch.device("cpu"),
) -> torch.Tensor:
    """Generate N points uniformly in the sphere(center, radius).
    Do so by generating points in a cube(center, radius) and discarding points outside the
    sphere.

    Args:
        center (torch.Tensor): Sphere's center coordinate tensor, of shape (3, ).
        radius (float): Sphere's radius.
        N (int, optional): Number of points generated. Defaults to 1.

    Returns:
        torch.Tensor: Generated point cloud, of shape (N, 3).
    """
    points = torch.zeros((N, 3)).to(device)
    selected = 0
    while selected < N:
        remainder = N - selected
        sel = min(remainder, N)
        point = (
            2 * (torch.rand(2 * sel, 3, device=points.device) - 0.5) * radius + center
        )
        point_distance = torch.norm(point - center, dim=1, keepdim=True)
        point = point[(point_distance <= radius).squeeze()]
        sel = min(sel, point.size(0))
        points[selected : (selected + sel)] = point[:sel]
        selected += sel
    return points


# _______________________________________________________________________________________________ #


def gaussian_digger(
    pointcloud: torch.Tensor, center: torch.Tensor, sigma: torch.Tensor
) -> torch.Tensor:
    """Randomly remove points from a point cloud following a gaussian probability with respect
    to a given center.

    Args:
        pointcloud (torch.Tensor): Point cloud tensor, of shape (N, 3).
        center (torch.Tensor): Gaussian distribution center. Tensor of shape (3, ).
        sigma (torch.Tensor): Gaussian distribution standard deviation. Tensor of shape (1, ).

    Returns:
        torch.Tensor: Randomly filtered point cloud, of shape (N', 3), where N' < N.
    """
    sigma_ = sigma.cpu().numpy()
    center_ = center.cpu().numpy()
    pc_ = pointcloud.detach().cpu().numpy()
    cov = sigma_**2 * np.eye(3)
    norm_factor = multivariate_normal.pdf(center_, mean=center_, cov=cov)
    thresholds = multivariate_normal.pdf(pc_, mean=center_, cov=cov) / norm_factor
    dices = np.random.uniform(low=0, high=1, size=(pointcloud.size(0)))
    choice_keep = dices >= thresholds
    return pointcloud[choice_keep]


# _______________________________________________________________________________________________ #


def non_uniform_density(
    pointcloud: torch.Tensor,
    nb_holes_min: int = 0,
    nb_holes_max: int = 10,
    mean_ratio: float = 0.5,
    std_ratio: float = 0.25,
    device: torch.device = torch.device("cpu"),
) -> torch.Tensor:
    """Simulate non uniform density of point cloud.
    Do so by randomly picking center and apply gaussian_digger onto them.

    Args:
        pointcloud (torch.Tensor): Point cloud tensor, of shape (N, 3).
        nb_holes_min (int, optional): Min number of intensity holes. Defaults to 0.
        nb_holes_max (int, optional): Max number of intensity holes. Defaults to 10.
        mean_ratio (float, optional): Holes will be from sphere following a gaussian proba
                                      of mean radius * mean_ratio. Defaults to 0.5.
        std_ratio (float, optional): Holes will be from sphere following a gaussian proba
                                      of std radius * mean_std. Defaults to 0.25.

    Returns:
        torch.Tensor: Randomly filtered point cloud, of shape (N', 3), where N' < N.
    """
    nb_holes = torch.randint(low=nb_holes_min, high=nb_holes_max, size=(1, 1)).squeeze()
    if nb_holes == 0:
        return pointcloud
    else:
        center, radius = get_center_and_radius(pointcloud)
        radius = torch.Tensor([radius]).to(device)
        hole_centers = uniform_in_sphere(center, radius, N=nb_holes, device=device)
        mean, std = radius * mean_ratio, radius * std_ratio
        if nb_holes == 1:
            sigma = torch.clamp(torch.normal(mean=mean, std=std), min=1e-3)
            return gaussian_digger(
                pointcloud, center=hole_centers.squeeze(), sigma=sigma
            )
        else:
            for hole_center in hole_centers:
                sigma = torch.clamp(torch.normal(mean=mean, std=std), min=1e-3)
                pointcloud = gaussian_digger(
                    pointcloud, center=hole_center, sigma=sigma
                )
            return pointcloud



import os
from abc import abstractmethod
from copy import deepcopy
from time import time
from typing import Tuple, Union

import numpy as np
import pandas as pd
import torch
from PIL import Image
from pyfigtree import figtree
from scipy.spatial import ConvexHull
from scipy.spatial.qhull import QhullError
from scipy.spatial.transform import Rotation
from torch.distributions.multivariate_normal import MultivariateNormal
from torch.nn import functional as F

from spfluo.picking.modules.pretraining.generate.data_generator import DataGenerator
from spfluo.segmentation.create_centriole import get_radius, simu_cylinder_rev
from spfluo.utils.loading import load_array, load_pointcloud
from spfluo.utils.volume import get_random_3d_vector


class SamplingStrategy:
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return torch.as_tensor(self.func(*self.args, **self.kwargs))


class Uniform3DAngle(SamplingStrategy):
    def __init__(self):
        super().__init__(self.func)

    def func(self):
        a = torch.rand(size=(3,))
        a[[0, 2]] *= 360
        a[1] *= 180
        return a


class Normal(SamplingStrategy):
    def __init__(self, size=(1,), loc=0, scale=1):
        super().__init__(self.func, size, loc, scale)

    def func(self, size, loc, scale):
        return torch.randn(size=size) * scale + loc


class DiscreteUniform(SamplingStrategy):
    def __init__(self, low, high):
        super().__init__(torch.randint, size=(1,), low=low, high=high)


class Uniform(SamplingStrategy):
    def __init__(self, low=0, high=1):
        super().__init__(self.func, low, high)

    def func(self, low, high):
        return torch.rand(1) * (high - low) + low


class Parameter(torch.nn.parameter.Parameter):
    def __new__(cls, arg, size=(1,), dtype=None, min=None, max=None):
        if isinstance(arg, SamplingStrategy):
            data = torch.empty(size, dtype=dtype)
        else:
            data = torch.tensor(arg)
        return super().__new__(cls, data=data, requires_grad=False)

    def __init__(self, arg, size=(1,), dtype=None, min=None, max=None):
        if isinstance(arg, SamplingStrategy):
            self.sampling_strategy = arg
        else:
            self.sampling_strategy = None
        self.clamp_params = {"min": min, "max": max}
        if min is not None and max is not None:
            self.do_clamp = True
        else:
            self.do_clamp = False

    def sample(self):
        if self.sampling_strategy is None:
            return self.data
        else:
            self.data = torch.as_tensor(
                self.sampling_strategy(), dtype=self.dtype, device=self.device
            )
            if self.do_clamp:
                self.data.clamp(**self.clamp_params)
            return self.data


class _AbstractBlock(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.requires_grad_(False)

    @abstractmethod
    def transform(self, data):
        pass

    def get_info(self):
        return dict(
            map(
                lambda x: (x[0], x[1].clone().float().squeeze().numpy()),
                self.named_parameters(recurse=False),
            )
        )

    def update_info(self, info):
        if self.__class__.__name__ not in info:
            info[self.__class__.__name__] = {}
        info[self.__class__.__name__].update(self.get_info())
        info[self.__class__.__name__].update({"applied": True})


class _AbstractAugment(_AbstractBlock):
    def __init__(self):
        super().__init__()

    def sample_params(self):
        return dict(
            map(lambda x: (x[0], x[1].sample()), self.named_parameters(recurse=False))
        )

    def forward(self, data):
        # t0 = time()
        self.sample_params()
        data = self.transform(data)
        # print(f"{self.__class__.__name__}: {time()-t0:.2f}s")
        return data


class _AbstractTransform(_AbstractBlock):
    def __init__(self):
        super().__init__()

    def forward(self, data):
        return self.transform(data)


class Sometimes(torch.nn.Module):
    def __init__(self, block: _AbstractBlock, p: float):
        super().__init__()
        self.block = block
        self.block_name = self.block.__class__.__name__
        self.p = p

    def forward(self, data):
        transformation_applied = torch.rand(1) < self.p
        if transformation_applied:
            data = self.block(data)
            info = data[-1]
        else:
            info = data[-1]
            self.block.update_info(info)
            for k in info[self.block_name]:
                info[self.block_name][k] += np.nan
            info[self.block_name]["applied"] = False
        return data


class _Generator(torch.nn.Sequential):
    def __init__(self):
        self.blocks = [
            CreatePointcloud,
            Rotate,
            Shrink,
            MakeHoles,
            Translate,
            DropPointcloud,
            Voxelise,
            CreateMask,
            AddRandomParticle,
            AddOutlierClusters,
            ApplyPSF,
            AddGaussianNoise,
            AddPoissonNoise,
        ]
        self.params = {c.__name__: {"p": 1, "args": {}} for c in self.blocks}

    def build_module(self):
        super().__init__(*self.build_pipeline(self.blocks, self.params))
        groundtruth_gen = torch.nn.Sequential(
            *self.build_pipeline(
                [CreatePointcloud, Voxelise],
                {
                    "CreatePointcloud": self.params["CreatePointcloud"],
                    "Voxelise": self.params["Voxelise"],
                },
            )
        )
        self.groundtruth = groundtruth_gen({})

    def build_pipeline(self, blocks, params):
        initialized_blocks = []
        for b_class in blocks:
            b = b_class(**params[b_class.__name__]["args"])
            p = params[b_class.__name__]["p"]
            if p < 1:
                initialized_blocks.append(Sometimes(b, p))
            else:
                initialized_blocks.append(b)
        return initialized_blocks

    def get_groundtruth(self):
        return self.groundtruth[0], self.groundtruth[2]


class _RandomParticle(_Generator):
    def __init__(self):
        super().__init__()

        self.params["CreatePointcloud"]["args"] = dict(
            a=-0.5, b=5.4, c=103.253, r=15, L=600, s=15, angle=110, N=1000
        )
        self.params["Rotate"]["args"]["rotation_angles"] = Uniform3DAngle()
        self.params["Shrink"]["args"]["factor"] = Normal(loc=1, scale=0.1)
        self.params["MakeHoles"]["args"] = dict(
            nb_holes=DiscreteUniform(0, 10), mean_ratio=0.5, std_ratio=0.25
        )
        self.params["Translate"]["args"]["tvec_norm"] = Normal(loc=100, scale=10)
        self.params["DropPointcloud"]["args"]["drop"] = True
        self.params["Voxelise"]["args"] = dict(
            center=(0, 0, 0), img_shape=(64, 64, 64), resolution=20
        )

        self.params["Shrink"]["p"] = 0.5
        self.params["MakeHoles"]["p"] = 0.5
        self.params["Translate"]["p"] = 0.5

        self.blocks = self.blocks[:8]
        self.blocks.pop(5)


class _DefaultGenerator(_Generator):
    def __init__(self):
        super().__init__()

        self.params["CreatePointcloud"]["args"] = dict(
            a=-0.5, b=5.4, c=103.253, r=15, L=600, s=15, angle=110, N=1000
        )
        self.params["Rotate"]["args"]["rotation_angles"] = Uniform3DAngle()
        self.params["Shrink"]["args"]["factor"] = Normal(loc=0.6, scale=0.01)
        self.params["MakeHoles"]["args"] = dict(
            nb_holes=DiscreteUniform(0, 10), mean_ratio=0.5, std_ratio=0.25
        )
        self.params["Translate"]["args"]["tvec_norm"] = Normal(loc=100, scale=10)
        self.params["DropPointcloud"]["args"]["drop"] = True
        self.params["Voxelise"]["args"] = dict(
            center=(0, 0, 0), img_shape=(64, 64, 64), resolution=20
        )
        rp = _RandomParticle()
        rp.build_module()
        self.params["AddRandomParticle"]["args"] = dict(
            create_particle=rp, distance=Normal(loc=5, scale=2), size=(64, 64, 64)
        )
        self.params["AddOutlierClusters"]["args"]["nb_clusters"] = DiscreteUniform(
            0, 10
        )
        self.params["AddGaussianNoise"]["args"] = dict(
            target_snr=Normal(loc=15, scale=2), gaussian_noise_mean=Uniform()
        )
        self.params["ApplyPSF"]["args"] = dict(
            cov_coef_sampling_strategy=Normal(loc=2, scale=1),
            anisotropy_sampling_strategy=Normal(loc=2, scale=0.5),
        )

        # set default prob
        self.params["Shrink"]["p"] = 0.7
        self.params["MakeHoles"]["p"] = 0.5
        self.params["Translate"]["p"] = 0.7
        self.params["DropPointcloud"]["p"] = 0.1
        self.params["AddRandomParticle"]["p"] = 0.7
        self.params["AddOutlierClusters"]["p"] = 0.8
        self.params["AddGaussianNoise"]["p"] = 0.6
        self.params["AddPoissonNoise"]["p"] = 0.6
        self.params["ApplyPSF"]["p"] = 1


class NoAugGenerator(_DefaultGenerator):
    def __init__(self, psf_coef=(0.1, 0.1, 0.1)):
        super().__init__()
        aug_blocks = [
            "Shrink",
            "Translate",
            "MakeHoles",
            "DropPointcloud",
            "AddRandomParticle",
            "AddOutlierClusters",
            "AddGaussianNoise",
            "AddPoissonNoise",
        ]
        for b in aug_blocks:
            self.params[b]["p"] = 0

        self.params["ApplyPSF"]["args"].update(
            dict(
                cov=psf_coef,
                cov_coef_sampling_strategy=None,
                anisotropy_sampling_strategy=None,
            )
        )

        self.build_module()


class SingleParticleGenerator_translation_holes(_DefaultGenerator):
    def __init__(
        self,
        image_shape=(32, 32, 32),
        voxelise_res=30,
        psf_coef=(0.1, 0.1, 0.1),
        nb_holes=10,
        tvec_norm=Normal(loc=100, scale=10),
    ):
        super().__init__()
        aug_blocks = [
            "Shrink",
            "DropPointcloud",
            "AddRandomParticle",
            "AddOutlierClusters",
            "AddGaussianNoise",
            "AddPoissonNoise",
        ]
        for b in aug_blocks:
            self.params[b]["p"] = 0

        self.params["Voxelise"]["args"]["resolution"] = voxelise_res

        self.params["Translate"]["args"]["tvec_norm"] = tvec_norm
        if nb_holes == 0:
            self.params["MakeHoles"]["args"] = dict(
                nb_holes=0, mean_ratio=0.5, std_ratio=0.25
            )
        else:
            self.params["MakeHoles"]["args"] = dict(
                nb_holes=DiscreteUniform(0, nb_holes), mean_ratio=0.5, std_ratio=0.25
            )
        self.params["Voxelise"]["args"].update({"img_shape": image_shape})
        self.params["Translate"]["p"] = 1
        self.params["MakeHoles"]["p"] = 1

        self.params["ApplyPSF"]["args"].update(
            dict(
                cov=psf_coef,
                cov_coef_sampling_strategy=None,
                anisotropy_sampling_strategy=None,
            )
        )

        self.build_module()


class RandomParticle_variableSize(_RandomParticle):
    def __init__(self, low, high=None):
        super().__init__()

        if high is None:
            high = low
        self.params["CreatePointcloud"]["args"].update(
            dict(
                L=Uniform(low, high),
            )
        )
        self.params["Shrink"]["args"].update(
            dict(
                factor=1.0,
            )
        )
        self.build_module()


class Generator_variableSize(_DefaultGenerator):
    def __init__(self, low, high=None):
        super().__init__()

        if high is None:
            high = low

        self.params["CreatePointcloud"]["args"].update(
            dict(
                L=Uniform(low, high),
            )
        )
        self.params["Shrink"]["args"].update(
            dict(
                factor=1.0,
            )
        )
        self.params["AddRandomParticle"]["args"].update(
            dict(create_particle=RandomParticle_variableSize(low, high))
        )

        self.build_module()


class ISIMParticle(_RandomParticle):
    def __init__(self):
        super().__init__()

        # Change params
        self.params["CreatePointcloud"]["args"].update(
            dict(
                a=Uniform(-0.535, 0),
                b=Uniform(5.4, 6),
                c=Uniform(103.253, 105),
                r=Uniform(10, 20),
                L=900,
                s=Uniform(10, 20),
                angle=Uniform(0, 180),
                N=1000,
            )
        )

        self.params["Voxelise"]["args"].update(
            {"img_shape": (64, 64, 64), "resolution": 20}
        )

        self.params["Shrink"]["args"]["factor"] = Normal(loc=1, scale=0.1)

        self.params["MakeHoles"]["args"] = dict(
            nb_holes=DiscreteUniform(100, 200), mean_ratio=0.1, std_ratio=0.02
        )

        self.build_module()


class ISIMGenerator(_DefaultGenerator):
    def __init__(self):
        super().__init__()

        # Change pointcloud loading
        # blocks[0] = CreatePointcloud
        # params.update({CreatePointcloud.__name__: {'p': 1, 'args': {}}})

        # Change params
        self.params["CreatePointcloud"]["args"].update(
            dict(
                a=Uniform(-0.535, 0),
                b=Uniform(5.4, 6),
                c=Uniform(103.253, 105),
                r=Uniform(10, 20),
                L=900,
                s=Uniform(10, 20),
                angle=Uniform(0, 180),
                N=1000,
            )
        )

        self.params["Shrink"]["args"]["factor"] = Normal(loc=1, scale=0.1)

        self.params["MakeHoles"]["args"] = dict(
            nb_holes=DiscreteUniform(100, 200), mean_ratio=0.1, std_ratio=0.02
        )

        self.params["Voxelise"]["args"].update(
            {"img_shape": (64, 64, 64), "resolution": 20}
        )

        self.params["AddRandomParticle"]["args"].update(
            dict(
                distance=Normal(loc=5, scale=2),
                size=(64, 64, 64),
                create_particle=ISIMParticle(),
            )
        )

        self.params["AddGaussianNoise"]["args"] = dict(
            target_snr=Normal(loc=20, scale=2),
            gaussian_noise_mean=Uniform(low=0.001, high=0.01),
        )

        self.params["ApplyPSF"]["args"] = dict(
            psf_path="/data/plumail/real_data/inputs/ISIM/psf.tif"
        )

        self.params["AddOutlierClusters"]["args"] = dict(
            nb_clusters=DiscreteUniform(100, 200),
            radius_ranges=((1, 6), (1, 6), (1, 6)),
            nb_points_range=(0, 20),
            intensity_range=(0.1, 0.5),
        )

        self.params["Shrink"]["p"] = 0.7
        self.params["MakeHoles"]["p"] = 1.0
        self.params["Translate"]["p"] = 0.8
        self.params["DropPointcloud"]["p"] = 0.0
        self.params["AddRandomParticle"]["p"] = 1.0
        self.params["AddOutlierClusters"]["p"] = 1.0
        self.params["AddGaussianNoise"]["p"] = 1.0
        self.params["AddPoissonNoise"]["p"] = 1.0
        self.params["ApplyPSF"]["p"] = 1

        self.build_module()


class Shrink(_AbstractAugment):
    def __init__(self, factor=1):
        super().__init__()
        self.factor = Parameter(factor, min=0.1)

    def transform(self, data):
        pointcloud, true_pointcloud, info = data
        self.update_info(info)
        if self.factor != 1:
            m = true_pointcloud.mean(dim=0)

            def shrink(x):
                return m + (x - m) / self.factor

            pointcloud = shrink(pointcloud)
            true_pointcloud = shrink(true_pointcloud)
        return pointcloud, true_pointcloud, info


class DropPointcloud(_AbstractAugment):
    def __init__(self, drop=True):
        super().__init__()
        self.empty_pointcloud = torch.Tensor([[], [], []]).permute(1, 0)
        self.pointcloud_dropped = Parameter(drop, dtype=torch.bool)

    def transform(self, data):
        pointcloud, true_pointcloud, info = data
        self.update_info(info)
        if self.pointcloud_dropped == 1:
            return self.empty_pointcloud.clone(), self.empty_pointcloud.clone(), info
        else:
            return pointcloud, true_pointcloud, info


class Translate(_AbstractAugment):
    def __init__(self, tvec_norm=0):
        super().__init__()
        if callable(tvec_norm):
            tvec_sampling_strategy = SamplingStrategy(
                lambda: get_random_3d_vector(tvec_norm())
            )
        else:
            tvec_sampling_strategy = SamplingStrategy(
                lambda: get_random_3d_vector(tvec_norm)
            )
        self.tvec = Parameter(tvec_sampling_strategy, size=(3,))

    def transform(self, data):
        pointcloud, true_pointcloud, info = data
        self.update_info(info)
        pointcloud += self.tvec
        true_pointcloud += self.tvec
        return pointcloud, true_pointcloud, info


class Rotate(_AbstractAugment):
    def __init__(self, rotation_angles=Uniform3DAngle()):
        super().__init__()
        self.rotation_angles = Parameter(rotation_angles, size=(3,))

    def transform(self, data):
        pointcloud, true_pointcloud, info = data
        self.update_info(info)

        rot = Rotation.from_euler(
            "ZXZ", self.rotation_angles.cpu().numpy(), degrees=True
        )
        rot_mat = torch.as_tensor(
            rot.as_matrix(), dtype=pointcloud.dtype, device=pointcloud.device
        )
        pointcloud = (rot_mat @ pointcloud[:, [2, 1, 0]].T).T
        true_pointcloud = (rot_mat @ true_pointcloud[:, [2, 1, 0]].T).T

        return pointcloud[:, [2, 1, 0]], true_pointcloud[:, [2, 1, 0]], info


class MakeHoles(_AbstractAugment):
    def __init__(self, nb_holes=0, mean_ratio=0, std_ratio=0):
        super().__init__()
        self.nb_holes = Parameter(nb_holes)
        self.mean_ratio = Parameter(mean_ratio)
        self.std_ratio = Parameter(std_ratio)

    def get_center_and_radius(
        self, pointcloud: torch.Tensor
    ) -> Tuple[torch.Tensor, float]:
        """Returns the center and the radius of a given point cloud.
        The center is the mean coordinate.
        The radius is given by the max distance between
        the center and any point from the point cloud.

        Args:
            pointcloud (torch.Tensor): Point cloud tensor, of shape (N, 3).

        Returns:
            Tuple[torch.Tensor, float]:
                * The center coordinates (x, y, z), of shape (3, ).
                * The radius as a single float.
        """
        center = pointcloud.mean(dim=0)
        distances = torch.norm(pointcloud - center, dim=1)
        return center, distances.max()

    def uniform_in_sphere(
        self, center: torch.Tensor, radius: float, N: int = 1
    ) -> torch.Tensor:
        """Generate N points uniformly in the sphere(center, radius).
        Do so by generating points in a cube(center, radius) and discarding points
        outside the sphere.

        Args:
            center (torch.Tensor): Sphere's center coordinate tensor, of shape (3, ).
            radius (float): Sphere's radius.
            N (int, optional): Number of points generated. Defaults to 1.

        Returns:
            torch.Tensor: Generated point cloud, of shape (N, 3).
        """
        N = int(N)
        points = center.new_zeros((N, 3))
        selected = 0
        while selected < N:
            remainder = N - selected
            sel = min(remainder, N)
            point = (
                2 * (torch.rand(2 * sel, 3, device=points.device) - 0.5) * radius
                + center
            )
            point_distance = torch.norm(point - center, dim=1, keepdim=True)
            point = point[(point_distance <= radius).squeeze()]
            sel = min(sel, point.size(0))
            points[selected : (selected + sel)] = point[:sel]
            selected += sel
        return points

    def get_params(self, pointcloud):
        center, radius = self.get_center_and_radius(pointcloud)
        holes_centers = self.uniform_in_sphere(center, radius, N=self.nb_holes)
        mean, std = radius * self.mean_ratio, radius * self.std_ratio
        return holes_centers, mean, std

    def transform(self, data: Tuple[torch.Tensor]) -> torch.Tensor:
        """Simulate non uniform density of point cloud.
        Do so by randomly picking center and apply gaussian_digger onto them.

        Args:
            pointcloud (torch.Tensor): Point cloud tensor, of shape (N, 3).
            nb_holes_min (int, optional): Min number of intensity holes.
                Defaults to 0.
            nb_holes_max (int, optional): Max number of intensity holes.
                Defaults to 10.
            mean_ratio (float, optional): Holes will be from sphere following a
                gaussian proba of mean radius * mean_ratio. Defaults to 0.5.
            std_ratio (float, optional): Holes will be from sphere following a gaussia
                proba of std radius * mean_std. Defaults to 0.25.

        Returns:
            torch.Tensor: Randomly filtered point cloud, of shape (N', 3), where N' < N.
        """
        pointcloud_noisy, true_pointcloud, info = data
        self.update_info(info)
        hole_centers, mean_scaled, std_scaled = self.get_params(pointcloud_noisy)
        nb_holes = len(hole_centers)
        if nb_holes > 0:
            sigma = torch.clamp(
                mean_scaled
                + torch.randn(size=(nb_holes,), device=std_scaled.device) * std_scaled,
                min=1e-3,
            )
            cov = (sigma[:, None, None] ** 2) * torch.eye(3, device=sigma.device)
            holes_distributions = MultivariateNormal(
                hole_centers, covariance_matrix=cov
            )  # nb_holes multivariate normal distributions
            prob = torch.exp(
                holes_distributions.log_prob(pointcloud_noisy[:, None])
            ) / torch.exp(
                holes_distributions.log_prob(hole_centers)
            )  # shape (N, nb_holes)
            dices = torch.rand(prob.size(), device=prob.device)
            keep = (dices >= prob).prod(dim=1).type(torch.bool)
            return pointcloud_noisy[keep], true_pointcloud, info
        else:
            return pointcloud_noisy, true_pointcloud, info


class Voxelise(_AbstractTransform):
    def __init__(
        self,
        center=(0, 0, 0),
        img_shape=(65, 65, 65),
        resolution=30,
        bandwidth=10,
        epsilon=1e-3,
        device=torch.device("cpu"),
        dtype=torch.float32,
    ):
        super().__init__()
        self.img_shape = Parameter(img_shape)
        self.bandwidth = Parameter(bandwidth)
        self.epsilon = Parameter(epsilon)
        self.center = Parameter(center)
        self.resolution = Parameter(resolution)

        self.device = device
        self.dtype = dtype

        start = (
            torch.as_tensor(center).cpu()
            - resolution * torch.as_tensor(img_shape).cpu() / 2
        )
        end = start + resolution * torch.as_tensor(img_shape).cpu()

        target_points = torch.meshgrid(
            *[
                torch.linspace(st, e, sh, device=self.device)
                for st, e, sh in zip(start, end, img_shape)
            ],
            indexing="ij",
        )
        self.target_points = torch.stack(target_points, dim=0)

    @staticmethod
    def get_FOV(
        pointcloud: Union[torch.Tensor, np.ndarray],
    ) -> Tuple[Tuple[float, float], ...]:
        """Get the Field Of View of a point cloud.

        Args:
            pointcloud (Union[torch.Tensor, np.ndarray]): shape (N, 3).

        Returns:
            Tuple[float]: x_min, x_max, y_min, y_max, z_min, z_max.
        """
        return tuple(
            [
                (pointcloud[:, i].min(), pointcloud[:, i].max())
                for i in range(pointcloud.shape[1])
            ]
        )

    def get_densities(self, pointcloud):
        if self.device.type == "cuda":
            return (
                1.0
                / (2.0 * torch.pi * self.bandwidth**2)
                * torch.sum(
                    torch.exp(
                        -(
                            (
                                pointcloud[:, :, None, None, None]
                                - self.target_points[None].to(pointcloud.device)
                            )
                            ** 2
                        ).sum(dim=1)
                        / (2 * self.bandwidth.to(pointcloud.device) ** 2)
                    ),
                    dim=0,
                )
            )
        else:  # use figtree
            weights = np.ones((len(pointcloud),))
            figtree_kwargs = {
                "bandwidth": self.bandwidth.cpu().numpy(),
                "epsilon": self.epsilon.cpu().numpy(),
            }
            if pointcloud.size(0) > 0:
                densities = figtree(
                    pointcloud.cpu().numpy().astype(np.float64),
                    self.target_points.cpu()
                    .numpy()
                    .astype(np.float64)
                    .transpose(1, 2, 3, 0)
                    .reshape(-1, 3),
                    weights,
                    **figtree_kwargs,
                )
                div = densities.max() - densities.min()
                if div > 0:
                    densities = (densities - densities.min()) / div
            else:
                densities = np.zeros(self.target_points.shape[1:])
            densities = torch.as_tensor(
                densities.reshape(self.target_points.shape[1:]),
                device=pointcloud.device,
                dtype=torch.float64,
            ).type(self.dtype)
            return densities

    def transform(self, data):
        pointcloud, true_pointcloud, info = data
        self.update_info(info)
        if "Translate" in "info" and info["Translate"]["applied"]:
            info["Translate"]["tvec"] -= self.center.cpu().numpy()
            info["Translate"]["tvec"] /= self.resolution.cpu().numpy()
        densities = self.get_densities(pointcloud)
        true_densities = self.get_densities(true_pointcloud)

        return densities, true_densities, info


class ApplyPSF(_AbstractAugment):
    def __init__(
        self,
        cov=(0.1, 0.1, 0.1),
        cov_coef_sampling_strategy=None,
        anisotropy_sampling_strategy=None,
        psf_path=None,
    ):
        super().__init__()
        if psf_path is not None and os.path.exists(psf_path):
            self.psf_path = psf_path
            self.psf = torch.as_tensor(load_array(psf_path), device="cpu").type(
                torch.float32
            )
        else:
            self.psf_path = None
            self.psf = None
            if (
                cov_coef_sampling_strategy is not None
                and anisotropy_sampling_strategy is not None
            ):
                cov_sampling = SamplingStrategy(
                    lambda: self.sample_cov(
                        cov_coef_sampling_strategy, anisotropy_sampling_strategy
                    )
                )
                self.cov_coef = Parameter(cov_sampling, size=(3,), min=0.1)
            else:
                self.cov_coef = Parameter(cov, min=0.1)

        """if torch.cuda.is_available():
            self.device = 'cuda'
        else:
            self.device = 'cpu'"""
        self.device = "cpu"

    def sample_cov(self, cov_sampling_strategy, anisotropy_sampling_strategy):
        c = cov_sampling_strategy()
        a = max(1.0, anisotropy_sampling_strategy())
        return torch.tensor([a * c, c, c])

    def get_psf(self, dtype):
        cov = torch.eye(3, dtype=dtype, device=self.device)
        cov[[0, 1, 2], [0, 1, 2]] = self.cov_coef.type(dtype).to(self.device) ** 2
        return torch.as_tensor(
            DataGenerator.make_psf(
                tuple((cov[[0, 1, 2], [0, 1, 2]] ** 0.5).cpu().numpy()), np.float64
            ),
            device=self.device,
            dtype=dtype,
        )

    def transform(self, data):
        img, mask, info = data
        if self.psf_path is None:
            self.update_info(info)
            psf = self.get_psf(img.dtype)
        else:
            psf = self.psf
            info["ApplyPSF"] = {"applied": True, "psf_path": self.psf_path}

        info["ApplyPSF"]["psf"] = psf
        img = F.conv3d(
            img[None, None].to(self.device), psf[None, None], padding="same"
        )[0, 0].to(img.device)
        # div = img.max() - img.min()
        # if div > 0:
        #    img = (img - img.min()) / div
        return img, mask, info


class CreateMask(_AbstractTransform):
    def __init__(self):
        super().__init__()

    def transform(self, data: Tuple[torch.Tensor]) -> Tuple[torch.Tensor]:
        img, true_img, info = data
        im = true_img.cpu().numpy()
        mask = np.zeros_like(im).astype(int)
        grid = np.indices(mask[0].shape).reshape(2, -1)
        for i in range(mask.shape[0]):
            ind = (im[i] > 0).nonzero()
            if len(ind[0]) > 0:
                try:
                    hull = ConvexHull(np.stack(ind, axis=1))
                    grid_ = np.concatenate((grid, np.ones((1, grid.shape[1]))), axis=0)
                    inside = (hull.equations @ grid_ < -1e-9).all(axis=0)
                    inside_points = grid[:, inside]
                    mask[i, inside_points[0], inside_points[1]] = 1
                except QhullError:
                    pass

        self.update_info(info)
        mask = torch.as_tensor(mask, dtype=torch.int16, device=img.device)
        return img, mask, info


class AddRandomParticle(_AbstractAugment):
    def __init__(self, create_particle=None, distance=0, size=(64, 64, 64)):
        super().__init__()

        self.create_particle = create_particle
        self.distance = Parameter(distance, min=-min(size), max=min(size))

    def get_translation_vec(self, reference_mask, mask_to_move, direction, distance):
        D, H, W = reference_mask.size()
        mask_moved_padded = F.pad(mask_to_move, (W, W, H, H, D, D))
        ref_mask = reference_mask > 0
        upper_bound = min((D, H, W))
        lower_bound = 0
        intersection_dist = 0
        while np.abs(upper_bound - lower_bound) > 1:
            intersection_dist = (upper_bound + lower_bound) / 2
            tvec = torch.round(direction * intersection_dist).type(torch.int64)
            mask_moved = mask_moved_padded[
                D - tvec[0] : D - tvec[0] + D,
                H - tvec[1] : H - tvec[1] + H,
                W - tvec[2] : W - tvec[2] + W,
            ]
            intersect = (ref_mask * mask_moved).sum() > 0
            if intersect:
                lower_bound = intersection_dist
            else:
                upper_bound = intersection_dist
        return torch.round(direction * (intersection_dist + distance)).type(torch.int64)

    def translate(self, img, tvec):
        D, H, W = img.size()
        img_padded = F.pad(img, (W, W, H, H, D, D))
        return img_padded[
            D - tvec[0] : D - tvec[0] + D,
            H - tvec[1] : H - tvec[1] + H,
            W - tvec[2] : W - tvec[2] + W,
        ]

    def move(self, img_to_move, ref_mask, mask_to_move, direction, distance):
        tvec = self.get_translation_vec(ref_mask, mask_to_move, direction, distance)
        img_moved = self.translate(img_to_move, tvec)
        mask_moved = self.translate(mask_to_move, tvec)
        return img_moved, mask_moved

    def transform(self, data):
        img, mask, info = data
        self.update_info(info)
        img2, mask2, info2 = self.create_particle({})
        assert "AddRandomParticle" not in info2

        if not info["DropPointcloud"]["applied"]:
            # Sample params
            random_direction = torch.as_tensor(get_random_3d_vector()).to(img.device)

            img2_moved, mask2_moved = self.move(
                img2, mask, mask2, random_direction, self.distance
            )

            img += img2_moved
            label_max = mask.max()
            mask[mask2_moved > 0] += label_max + 1
            mask[mask > label_max + 1] = -1  # overlapping between 2 particles
        else:
            info["AddRandomParticle"]["applied"] = False

        return img, mask, info


class AddOutlierClusters(_AbstractAugment):
    def __init__(
        self,
        nb_clusters=0,
        radius_ranges=((20, 80), (20, 80), (25, 80)),
        nb_points_range=(20, 80),
        intensity_range=(0, 1),
    ):
        super().__init__()
        self.nb_clusters = Parameter(nb_clusters)
        self.radius_ranges = radius_ranges
        self.nb_points_range = nb_points_range
        self.intensity_range = intensity_range

    def sample_outliers_clusters(self, img):
        if self.nb_clusters > 0:
            radius = torch.stack(
                [
                    torch.randint(r[0], r[1], size=(int(self.nb_clusters),))
                    for r in self.radius_ranges
                ],
                dim=1,
            )
            cov = radius[:, None] ** 0.5 * torch.eye(3)
            center = torch.as_tensor(
                [
                    [
                        torch.randint(0, img.shape[i], size=(1,))
                        for i in range(radius.size(1))
                    ]
                    for j in range(radius.size(0))
                ]
            ).type(cov.dtype)
            low, high = self.nb_points_range
            nb_points = torch.randint(low, high, size=(int(self.nb_clusters),))
            dist = MultivariateNormal(center, cov)
            outliers_clusters = dist.sample((max(nb_points),)).permute(1, 0, 2)
            outliers_clusters = torch.round(outliers_clusters).type(torch.int64)
            outliers_clusters = [
                cluster[:n] for cluster, n in zip(outliers_clusters, nb_points)
            ]
            masks = [
                ((cluster > 0) * (cluster < torch.as_tensor(img.shape)))
                .prod(dim=1)
                .type(torch.bool)
                for cluster in outliers_clusters
            ]
            outliers_clusters = [
                cluster[mask] for cluster, mask in zip(outliers_clusters, masks)
            ]
            return outliers_clusters
        else:
            return []

    def transform(self, data):
        img, mask, info = data
        self.update_info(info)
        outliers_clusters = self.sample_outliers_clusters(img)
        low, high = self.intensity_range
        for cluster in outliers_clusters:
            intensities = (high - low) * torch.rand(
                (len(cluster),), device=img.device, dtype=img.dtype
            ) + low
            img[cluster[:, 0], cluster[:, 1], cluster[:, 2]] = intensities

        return img, mask, info


class AddGaussianNoise(_AbstractAugment):
    def __init__(
        self,
        target_snr=100,
        gaussian_noise_mean=0,
    ):
        super().__init__()
        self.target_snr = Parameter(target_snr)
        self.gaussian_noise_mean = Parameter(gaussian_noise_mean)

    def transform(self, data):
        img, mask, info = data
        self.update_info(info)

        d = img.max() - img.min()
        img = (img - img.min()) / d if d > 0 else img - img.min()
        image_average_db = 10 * torch.log10(torch.clamp(img.mean(), 0.01))
        noise_average_db = image_average_db - self.target_snr
        noise_average = 10 ** (noise_average_db / 10)
        noise = torch.normal(
            float(self.gaussian_noise_mean), float(noise_average) ** 0.5, size=img.shape
        ).to(img.device)
        img += noise
        img = torch.clamp(img, 0, 1)

        return img, mask, info


class AddPoissonNoise(_AbstractAugment):
    def __init__(self):
        super().__init__()

    def transform(self, data):
        img, mask, info = data
        self.update_info(info)

        # Determine unique values in image & calculate the next power of two
        img = img.clamp(0, None)
        vals = torch.as_tensor(len(torch.unique(img)))
        vals = 2 ** torch.ceil(torch.log2(vals))

        # Generating noise for each unique value in image.
        img = torch.poisson(img * vals) / vals

        return img, mask, info


class LoadPointcloud(_AbstractTransform):
    def __init__(
        self, pointcloud_path="/data1/stage_Luc/inputs/sample_centriole_point_cloud.csv"
    ):
        super().__init__()
        self.pointcloud_path = pointcloud_path
        pointcloud = load_pointcloud(pointcloud_path)
        self.pointcloud = torch.as_tensor(pointcloud)[:, [2, 1, 0]]
        self.pointcloud -= self.pointcloud.mean(dim=0)

    def transform(self, data):
        info = data
        info["LoadPointcloud"] = {"applied": True}
        info["LoadPointcloud"]["pointcloud_path"] = self.pointcloud_path
        return self.pointcloud.clone(), self.pointcloud.clone(), info


class CreatePointcloud(_AbstractAugment):
    def __init__(self, a, b, c, r, L, s, angle, N, device=None, dtype=None):
        super().__init__()
        self.a = Parameter(a)
        self.b = Parameter(b)
        self.c = Parameter(c)
        self.r = Parameter(r)
        self.L = Parameter(L)
        self.s = Parameter(s)
        self.angle = Parameter(angle)
        self.N = N

        self.device = device
        self.dtype = dtype

    def create_centriole(self):
        x = np.arange(1, 14)
        pc = simu_cylinder_rev(
            get_radius(
                x, self.a.cpu().numpy(), self.b.cpu().numpy(), self.c.cpu().numpy()
            ),
            self.r.cpu().numpy(),
            self.L.cpu().numpy(),
            self.s.cpu().numpy(),
            self.N,
            self.angle.cpu().numpy(),
        )
        pc -= pc.mean(axis=0)
        return torch.as_tensor(pc, dtype=self.dtype, device=self.device)[:, [2, 1, 0]]

    def transform(self, data):
        info = data
        self.update_info(info)
        pointcloud = self.create_centriole()
        return pointcloud, pointcloud.clone(), info


class GeneratorDataset(torch.utils.data.Dataset):
    def __init__(self, generator, length):
        super().__init__()
        self.generator = generator
        self.length = length

    def __getitem__(self, idx):
        im, mask, info = self.generator({})
        im = im.unsqueeze(0)
        return im, mask, info

    def __len__(self):
        return self.length


def unbatch_dict(d):
    res = []
    for k in d:
        if isinstance(d[k], dict):
            r = unbatch_dict(d[k])
            for i in range(len(r)):
                if len(res) < i + 1:
                    res.append({})
                res[i][k] = r[i]
        else:
            for i in range(len(d[k])):
                if len(res) < i + 1:
                    res.append({})
                res[i][k] = d[k][i]
    return res


def infos_to_dataframe(infos):
    b = {}
    for k in infos:
        if isinstance(infos[k], dict):
            for k2 in infos[k]:
                if len(infos[k][k2].shape) == 1:
                    b[k + "_" + k2] = infos[k][k2]
                elif len(infos[k][k2].shape) == 2:
                    for i in range(infos[k][k2].shape[1]):
                        b[k + "_" + k2 + "_" + str(i)] = infos[k][k2][:, i]
        else:
            b[k] = infos[k]

    return pd.DataFrame(b)


class HardCopyDataset(torch.utils.data.Dataset):
    def __init__(self, dataset, **kwargs):
        loader = torch.utils.data.DataLoader(dataset, batch_size=1, **kwargs)
        self.data = [
            (deepcopy(im[0]), deepcopy(mask[0]), deepcopy(unbatch_dict(info)[0]))
            for im, mask, info in loader
        ]

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx]


class MultiEpochsDataLoader(torch.utils.data.DataLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._DataLoader__initialized = False
        self.batch_sampler = _RepeatSampler(self.batch_sampler)
        self._DataLoader__initialized = True
        self.iterator = super().__iter__()

    def __len__(self):
        return len(self.batch_sampler.sampler)

    def __iter__(self):
        for i in range(len(self)):
            yield next(self.iterator)


class _RepeatSampler(object):
    """Sampler that repeats forever.
    Args:
        sampler (Sampler)
    """

    def __init__(self, sampler):
        self.sampler = sampler

    def __iter__(self):
        while True:
            yield from iter(self.sampler)


if __name__ == "__main__":
    import napari
    from torch.utils.data import DataLoader

    pc_path = "/data1/stage_Luc/inputs/sample_centriole_point_cloud.csv"
    ds = GeneratorDataset(ISIMGenerator(), 1000)
    print(ds)
    data = ds[0]

    """
    data = None
    for m in ds.generate_image[:3]:
        data = m(data)
    
    for m in ds.generate_image[3]:
        t0 = time()
        data = m(data)
        print(m, time()-t0)
    """

    data_loader = DataLoader(ds, batch_size=8, num_workers=64, pin_memory=True)

    i = 0
    while 1:
        im, mask, info = ds[0]
        im = im.cpu().numpy()[0]
        mask = mask.cpu().numpy()
        # print(info)
        # viewer = napari.Viewer(ndisplay=3)
        # viewer.add_image(im, contrast_limits=(0,1))
        # viewer.add_labels(mask, visible=False)
        # napari.run()

        mask_ = np.zeros((mask.shape[1], mask.shape[2], 3), dtype=np.uint8)
        mask_[mask[mask.shape[0] // 2] == 1, :] = [0, 255, 0]
        mask_[mask[mask.shape[0] // 2] == 2, :] = [255, 0, 0]

        im = im[im.shape[0] // 2]
        im = (im - im.min()) / (im.max() - im.min())
        im = (im * 255).astype(np.uint8)
        im = Image.fromarray(im)
        mask = Image.fromarray(mask_)
        im.save(f"segmentation/data/generated/im{i}.png")
        mask.save(f"segmentation/data/generated/mask{i}.png")
        i += 1

    t0 = time()
    i = 0
    for im, mask, info in data_loader:
        print(i)
        im, mask = im.to("cuda"), mask.to("cuda")
        i += 1
    print(time() - t0)

    viewer = napari.Viewer(ndisplay=3)
    viewer.add_image(im.cpu().numpy())
    viewer.add_labels(mask.cpu().numpy())
    napari.run()



import atexit
import csv
import itertools
import os
from typing import TYPE_CHECKING, Tuple

import napari
import numpy as np
import tifffile
from napari.utils.events import Event
from napari_bbox import BoundingBoxLayer
from ome_types import from_xml
from vispy.color import Colormap

from spfluo.utils.loading import get_data_from_ome_tiff
from spfluo.visualisation.multiple_viewer_widget import add_orthoviewer_widget, init_qt

if TYPE_CHECKING:
    from napari.components import ViewerModel


def annotate(
    im_path: str,
    output_path: str,
    spacing: Tuple[float, float, float] = (1, 1, 1),
    save: bool = True,
):
    """
    Inputs:
        im_path: the 3D image to read.
        output_path: the path of the output CSV. If a CSV is already present,
            read it and display.
        size: is in pixels (cannot do it in real size)
        spacing: is the size of the pixels in um, ZYX order.
    Ouput:
        The CSV written in output_path is in real coordinates, in um.
        The last column is the diameter of the circle in um.
    """
    # correct spacing, see https://github.com/napari/napari/issues/6627
    spacing_normalized = np.asarray(spacing) / np.min(spacing)
    edge_width = max(
        (
            0.4  # 0.4um: length of a centriole
            / 10  # 1/10th of the length of a centriole
            / spacing[1]  # in pixel space
        ),
        2,
    )

    init_qt()

    bbox_layer = BoundingBoxLayer(
        ndim=3,
        edge_color=[255, 0, 255, 255],
        face_color=[0, 0, 0, 0],
        scale=spacing_normalized,
        name="Picking",
    )

    if os.path.exists(output_path):
        with open(output_path, "r", newline="") as csvfile:
            # read csv file into array
            data = []
            reader = csv.reader(csvfile)
            try:
                next(reader)
            except StopIteration:  # file is empty
                reader = []
            for p in reader:
                p1 = bbox_layer.world_to_data(
                    np.asarray(
                        (float(p[0 * 3 + 1]), float(p[0 * 3 + 2]), float(p[0 * 3 + 3]))
                    )
                    / np.min(spacing)
                )
                p2 = bbox_layer.world_to_data(
                    np.asarray(
                        (float(p[1 * 3 + 1]), float(p[1 * 3 + 2]), float(p[1 * 3 + 3]))
                    )
                    / np.min(spacing)
                )
                data.append(
                    list(
                        itertools.product(
                            (p1[0], p2[0]), (p1[1], p2[1]), (p1[2], p2[2])
                        )
                    )
                )
            data = np.asarray(data)
        bbox_layer.data = data

    view = napari.Viewer()
    view, dock_widget, cross = add_orthoviewer_widget(view)

    with tifffile.TiffFile(im_path, is_ome=True) as tif:
        im = get_data_from_ome_tiff(tif, 0, order="CZYX")
        ome = from_xml(tif.ome_metadata)
    assert len(ome.images) == 1
    if im.shape[0] == len(ome.images[0].pixels.channels):
        # define colormap
        cm = [
            Colormap(
                [[0.0, 0.0, 0.0], [float(x) / 255.0 for x in c.color.as_rgb_tuple()]]
            )
            for c in ome.images[0].pixels.channels
        ]
    else:
        cm = None
    view.add_image(
        im,
        scale=spacing_normalized,
        channel_axis=0,
        colormap=cm,
        blending="additive",
    )

    viewers: list[ViewerModel] = [
        dock_widget.viewer,
        dock_widget.viewer_model1,
        dock_widget.viewer_model2,
    ]

    def update_viewer(layer: BoundingBoxLayer, event: Event):
        dragged = False
        yield
        while event.type == "mouse_move":
            dragged = True
            yield
        if len(layer.selected_data) > 0:
            idx_point = list(layer.selected_data)[0]
            if dragged:
                layer.data = layer.data  # seems stupid but very important
            try:
                bbox = tuple(layer.data[idx_point])
            except IndexError:
                return
            bbox_center = np.array(bbox).mean(axis=0)
            # update viewers
            viewers_not_under_mouse = [
                viewer for viewer in viewers if not viewer.mouse_over_canvas
            ]
            if (
                len(viewers_not_under_mouse) == 3
            ):  # if mouse is not over any viewer, the point size is being adjusted
                return
            for viewer in viewers_not_under_mouse:
                if len(viewer.dims.order) > 3:
                    order = [x - 1 for x in viewer.dims.order[1:]]
                else:
                    order = list(viewer.dims.order)
                pos_reordered = tuple(
                    np.array(bbox_layer.data_to_world(bbox_center))[order]
                )
                viewer.camera.center = pos_reordered

            if len(dock_widget.viewer.dims.order) > 3:
                range_ = dock_widget.viewer.dims.range[1:]
            else:
                range_ = dock_widget.viewer.dims.range
            dock_widget.viewer.dims.current_step = tuple(
                np.round(
                    [
                        max(min_, min(p, max_)) / step
                        for p, (min_, max_, step) in zip(
                            bbox_layer.data_to_world(bbox_center), range_, strict=True
                        )
                    ]
                ).astype(int)
            )

    view.add_layer(bbox_layer)
    for v in viewers:
        for layer in v.layers:
            if isinstance(layer, BoundingBoxLayer):
                layer.mouse_drag_callbacks.append(update_viewer)
                layer.edge_width = edge_width
    view.scale_bar.visible = True
    view.scale_bar.unit = "um"

    # Save annotations
    if save:
        f = open(output_path, "w")

        def save_annotations():
            # delete previous annotations
            f.seek(0)
            f.truncate()

            # write new annotations
            f.write(
                ",".join(
                    [
                        "index",
                        "axis-1",
                        "axis-2",
                        "axis-3",
                        "axis-1",
                        "axis-2",
                        "axis-3",
                    ]
                )
            )
            f.write("\n")
            for i, bbox in enumerate(bbox_layer.data):
                f.write(str(i) + ",")
                f.write(
                    ",".join(
                        map(
                            str,
                            np.asarray(bbox_layer.data_to_world(np.min(bbox, axis=0)))
                            * np.min(spacing),
                        )
                    )
                )
                f.write(",")
                f.write(
                    ",".join(
                        map(
                            str,
                            np.asarray(bbox_layer.data_to_world(np.max(bbox, axis=0)))
                            * np.min(spacing),
                        )
                    )
                )
                f.write("\n")
            f.tell()

        save_annotations()
        bbox_layer.events.data.connect(save_annotations)

        bbox_layer.mode = "ADD_BOUNDING_BOX"
        atexit.register(lambda: f.close())

    napari.run()



import csv
import io
import sys
import tempfile
import zipfile
from contextlib import contextmanager
from pathlib import Path
from typing import Dict

import requests
import tifffile

from spfluo.data.constants import ARCHIVE_NAME, REPO_TOKEN, SEAFILE_URL
from spfluo.data.upload import _file_hash
from spfluo.utils.array import numpy as np
from spfluo.utils.loading import read_poses
from spfluo.utils.read_save_files import read_image

if sys.version_info >= (3, 9):
    from importlib.resources import as_file, files

    @contextmanager
    def _get_data_dir():
        with as_file(files("spfluo").joinpath("data")) as data_dir:
            yield data_dir
else:

    @contextmanager
    def _get_data_dir():
        yield Path(__file__).parent


def _download_data(d: Path):
    headers = {
        "Authorization": f"Token {REPO_TOKEN}",
        "Accept": "application/json",
    }
    download_link = requests.get(
        f"{SEAFILE_URL}/api/v2.1/via-repo-token/download-link/?path={ARCHIVE_NAME}",
        headers=headers,
    )
    download_link.raise_for_status()
    download_link = download_link.text.strip('"')

    response = requests.get(download_link)
    if response.status_code == 200:
        # Write the content of the response to a file
        with tempfile.TemporaryFile() as fp:
            fp.write(response.content)
            fp.seek(0)
            with zipfile.ZipFile(io.BytesIO(fp.read()), "r") as zip_ref:
                zip_ref.extractall(d)
            return True
    return False


def _check_data(d: Path):
    registry = d / "registry.txt"
    for line in registry.read_text().strip().split("\n"):
        path, hash = line.split(" ")
        path = d / path
        if (not path.exists()) or _file_hash(path) != hash:
            return False
    return True


@contextmanager
def _get_data():
    with _get_data_dir() as data_dir:
        if not _check_data(data_dir):
            _download_data(data_dir)
            if not _check_data(data_dir):
                raise RuntimeError("Download failed")
        yield data_dir


def _fetch_generated_dataset(dataset_name: str) -> Dict[str, np.ndarray]:
    # Download if necessary
    with _get_data() as data_dir:
        # parse data
        root_dir = data_dir / "generated" / dataset_name
        poses_path = root_dir / "poses.csv"
        content = csv.reader(poses_path.read_text().split("\n"))
        next(content)  # skip header
        data = {}
        for row in content:
            if len(row) == 7:
                particle_path = (root_dir / row[0]).as_posix()
                data[row[0]] = {
                    "array": tifffile.imread(particle_path),
                    "rot": np.array(row[1:4], dtype=float),
                    "trans": np.array(row[4:7], dtype=float),
                }

        N = len(data)
        p0 = next(iter(data))
        D = data[p0]["array"].shape[0]
        dtype = data[p0]["array"].dtype
        volumes = np.empty((N, D, D, D), dtype=dtype)
        poses = np.empty((len(data), 6))
        for i, k in enumerate(data):
            p = data[k]
            rot = p["rot"]
            trans = p["trans"]
            poses[i, :3] = rot
            poses[i, 3:] = trans
            volumes[i] = p["array"]

        psf = tifffile.imread((root_dir / "psf.tiff").as_posix())
        gt = tifffile.imread((root_dir / "gt.tiff").as_posix())

        return {
            "volumes": volumes,
            "poses": poses,
            "psf": psf,
            "gt": gt,
            "rootdir": poses_path.parent.absolute(),
        }


def generated_isotropic():
    return _fetch_generated_dataset("isotropic-1.0")


def generated_anisotropic():
    return _fetch_generated_dataset("anisotropic-5.0-1.0-1.0")


def generated_with_translations():
    return _fetch_generated_dataset("with-translations")


def real_ab_initio_reconstruction():
    with _get_data_dir() as data_dir:
        recon = data_dir / "real" / "reconstruction.tiff"
        poses = data_dir / "real" / "poses.csv"
        poses, names = read_poses(str(poses))
        pose = poses[names.index("reconstruction.tiff")]
        recon = read_image(str(recon), dtype="float64", xp=np)
        return {"reconstruction": recon, "pose": pose}



import csv
import os
import shutil
from typing import List, Tuple

import numpy as np
import tifffile
from tqdm import tqdm

from ..training import make_U_mask
from ..utils import center_to_corners, load_annotations, load_array, summary

# +------------------------------------------------------------------------------------------+ #
# |                                        MAKE U MASKS                                      | #
# +------------------------------------------------------------------------------------------+ #


def make_U_masks(rootdir: str, extension: str = "npz") -> None:
    """While U masks can be made on the fly during training, it is possible to generate all
    of them upstream. This can save time when running multiples training on the same data.
    """
    print(f"Current directory: {rootdir}")
    # 1. Load images
    images_names = sorted(
        list(filter(lambda x: x.endswith(extension), os.listdir(rootdir)))
    )
    images_paths = sorted(list(map(lambda x: os.path.join(rootdir, x), images_names)))
    images = [load_array(path) for path in tqdm(images_paths, desc="Load images")]
    # 2. Make U masks
    U_masks = [make_U_mask(image) for image in tqdm(images, desc="Create masks")]
    # 3. Save U masks
    output_dir = os.path.join(rootdir, "U_masks")
    os.makedirs(output_dir, exist_ok=True)
    for name, u_mask in tqdm(
        zip(images_names, U_masks), total=len(U_masks), desc="Save masks"
    ):
        tifffile.imwrite(os.path.join(output_dir, name), u_mask)


# +------------------------------------------------------------------------------------------+ #
# |                  RANDOM (NEGATIVE) AND AROUND PARTICLES (POSITIVE) CROPS                 | #
# +------------------------------------------------------------------------------------------+ #

DEFAULT_NEGATIVE_CENTERS = 8


def generate_random_negative_centers(
    positive_centers: np.ndarray,
    crop_size: Tuple[int],
    shape: Tuple[int],
    pos_ratio: float,
) -> np.ndarray:
    num_negative_centers = int(len(positive_centers) / pos_ratio)
    if num_negative_centers < DEFAULT_NEGATIVE_CENTERS:
        num_negative_centers = DEFAULT_NEGATIVE_CENTERS
    negative_centers = np.array([0, 0, 0])
    while len(negative_centers) <= num_negative_centers:
        current_centers = np.vstack((negative_centers, positive_centers))
        border = np.array(crop_size) // 2
        low, high = border, np.array(shape) - border
        new_center = np.random.randint(low, high=high)
        center_tile = np.repeat(new_center[np.newaxis, :], len(current_centers), axis=0)
        distances = np.linalg.norm(center_tile - current_centers.astype(float), axis=1)
        if distances.min() > max(crop_size):
            negative_centers = np.vstack((negative_centers, new_center))
    return negative_centers[1:]


def random_shift_center(
    center: Tuple[int],
    crop_size: Tuple[int],
    max_size: Tuple[int],
    margin: Tuple[int],
) -> Tuple[int]:
    random_shifts = []
    for s, c, m, M in zip(crop_size, center, margin, max_size):
        inner_range = s // 2 - M // 2 - m
        shift_max = min(inner_range, M - c - s // 2)
        shift_min = max(-inner_range, -c + s // 2)
        if shift_min >= shift_max:
            random_shifts.append(0)
        else:
            random_shifts.append(np.random.randint(shift_min, shift_max))
    new_center = [c + s for c, s in zip(center, random_shifts)]
    return new_center


def reframe_corners_if_needed(
    corners: Tuple[int], crop_size: Tuple[int], max_size: Tuple[int]
) -> Tuple[int]:
    d, h, w = crop_size
    D, H, W = max_size
    x_min, y_min, z_min, x_max, y_max, z_max = corners
    z_min, y_min, x_min = max(0, z_min), max(0, y_min), max(0, x_min)
    z_max, y_max, x_max = z_min + d, y_min + h, x_min + w
    z_max, y_max, x_max = min(D, z_max), min(H, y_max), min(W, x_max)
    z_min, y_min, x_min = z_max - d, y_max - h, x_max - w
    return x_min, y_min, z_min, x_max, y_max, z_max


def crop_one_particle(
    image: np.ndarray,
    center: np.ndarray,
    crop_size: Tuple[int],
    max_size: Tuple[int],
    margin: Tuple[int],
    random_shift: bool,
) -> np.ndarray:
    if random_shift:
        center = random_shift_center(center, crop_size, max_size, margin)
    corners = center_to_corners(center, crop_size)
    corners = reframe_corners_if_needed(corners, crop_size, max_size)
    x_min, y_min, z_min, x_max, y_max, z_max = corners
    return image[z_min:z_max, y_min:y_max, x_min:x_max]


def save(
    image: np.ndarray,
    image_name: str,
    crop_index: str,
    output_dir: str,
    extension: str = "npz",
) -> None:
    image_prefix = os.path.splitext(image_name)[0]
    name = f"{image_prefix}_{crop_index}"
    output = os.path.join(output_dir, name)
    if extension == "npz":
        np.savez(output + ".npz", image=image)
    elif extension == "tiff" or extension == "tif":
        image = image.astype(float)
        image = (image - image.min()) / (image.max() - image.min())
        image = (image * 255).astype(np.uint8)
        tifffile.imwrite(output + ".tiff", image)


def crop(
    rootdir: str,
    output_dir: str,
    crop_size: Tuple[int],
    margin: Tuple[int],
    pos_ratio: float = 1.0,
    positive_only: bool = False,
    extension: str = "npz",
    csv_name: str = "train_coordinates",
) -> None:
    """Crop around particles and save cropped images with respect to a given ratio of positive
    samples.

    Args:
        rootdir (str): Path to npz containing 3d images and masks to be cropped and a csv with
                       particles center coordinates.
        crop_size (Optional, Tuple[int]): (depth, height, width).
        margin (Optional, Tuple[int]): When randomly shifting bbox, on each axis the shift
                                        range will be (size // 2 - max_particle_dim // 2 - margin),
                                        and when generating random negative centers, they will be
                                        at least margin away from image border.
        pos_ratio (Optional, float): Will save N positive crops and N / ratio negative crops.
                                     In [0, 1].
        positive_only (Optional, bool): Wether or not to generate positive crop only. Usefull for
                                        tilt learning. Default to False.
        csv_name (Optional, str): Name of the annotations csv to load.
                                 Defaults to 'train_coordinates'.
    """
    positive_dir = os.path.join(rootdir, output_dir, "positive")
    negative_dir = os.path.join(rootdir, output_dir, "negative")
    os.makedirs(positive_dir, exist_ok=True)
    os.makedirs(negative_dir, exist_ok=True)
    images_list = sorted(
        list(filter(lambda x: x.endswith(extension), os.listdir(rootdir)))
    )
    annotations = load_annotations(os.path.join(rootdir, f"{csv_name}.csv"))
    print(f"Current directory: {rootdir}")
    print(
        f"Found {len(images_list)} images with {len(annotations)} particles annotated."
    )
    if len(annotations) > 0:
        for image_name in tqdm(images_list):
            image = load_array(os.path.join(rootdir, image_name))
            centers = annotations[annotations[:, 0] == image_name, 2:].astype(float)
            indices = annotations[annotations[:, 0] == image_name, 1]
            centers = np.rint(centers).astype(int)
            for crop_index, center in zip(indices, centers):
                crop_args = [image, center, crop_size, image.shape, margin]
                cropped_image = crop_one_particle(
                    *crop_args, random_shift=False
                )  # was True
                save(cropped_image, image_name, crop_index, positive_dir, extension)
            if not positive_only:
                args = [centers, crop_size, image.shape, pos_ratio]
                negative_centers = generate_random_negative_centers(*args)
                for crop_index, center in enumerate(negative_centers):
                    crop_args = [image, center, crop_size, image.shape, margin]
                    cropped_image = crop_one_particle(*crop_args, random_shift=False)
                    save(cropped_image, image_name, crop_index, negative_dir, extension)


# +------------------------------------------------------------------------------------------+ #
# |                              TRAIN/VAL/TEST SPLITS AND PATHS                             | #
# +------------------------------------------------------------------------------------------+ #


def train_test_split(
    rootdir: str, images: List[str], image_format: str = "npz", split: float = 0.8
) -> None:
    """Creates train/ and test/ rootdir's subfolders and put into them splitted images
        and annotations.

    Args:
        rootdir (str): Folder containing the images to split and the annotations csv.
        image_format (str, optional): Image file format. Defaults to 'npz'.
        split (float, optional): Ratio len(train) / len(test). Defaults to 0.8.
    """
    train_length = int(split * len(images))
    train_images, test_images = images[:train_length], images[train_length:]
    train_dir, test_dir = os.path.join(rootdir, "train"), os.path.join(rootdir, "test")
    os.makedirs(train_dir, exist_ok=True)
    os.makedirs(test_dir, exist_ok=True)
    for image in train_images:
        shutil.copyfile(os.path.join(rootdir, image), os.path.join(train_dir, image))
    for image in test_images:
        shutil.copyfile(os.path.join(rootdir, image), os.path.join(test_dir, image))
    annotations = load_annotations(os.path.join(rootdir, "coordinates.csv"))
    train_annotations = []
    for image in train_images:
        train_annotations.extend(annotations[annotations[:, 0] == image])
    train_annotations = np.array(train_annotations, dtype=object)
    csv_path = os.path.join(train_dir, "train_coordinates.csv")
    with open(csv_path, "w", newline="") as f:
        csv.writer(f).writerows(train_annotations)
    test_annotations = []
    for image in test_images:
        test_annotations.extend(annotations[annotations[:, 0] == image])
    test_annotations = np.array(test_annotations, dtype=object)
    csv_path = os.path.join(test_dir, "test_coordinates.csv")
    with open(csv_path, "w", newline="") as f:
        csv.writer(f).writerows(test_annotations)

    return train_images


def train_val_split(
    traindir: str, images: List[str], image_format: str = "npz", split: float = 0.8
) -> None:
    """After a train/test split has occured, an other train/val split is required inside the
    train subfolder. This could be done on the fly during training but hard separating the files
    seems cleaner and force a constant validation set regardless of the training config.

    Args:
        traindir (str): The train subfolder.
        image_format (str, optional): Used to detect images inside the given folder.
                                      Defaults to 'npz'.
        split (float, optional): Proportion of images to keep for training. Defaults to 0.8.
    """
    train_length = int(split * len(images))
    train_images, val_images = images[:train_length], images[train_length:]
    parent_dir = os.path.dirname(os.path.abspath(traindir))
    val_dir = os.path.join(parent_dir, "val")
    os.makedirs(val_dir, exist_ok=True)
    for image in val_images:
        os.replace(os.path.join(traindir, image), os.path.join(val_dir, image))
    annotations = load_annotations(os.path.join(traindir, "train_coordinates.csv"))
    train_annotations = []
    for image in train_images:
        train_annotations.extend(annotations[annotations[:, 0] == image])
    train_annotations = np.array(train_annotations, dtype=object)
    csv_path = os.path.join(traindir, "train_coordinates.csv")
    with open(csv_path, "w", newline="") as f:
        csv.writer(f).writerows(train_annotations)
    val_annotations = []
    for image in val_images:
        val_annotations.extend(annotations[annotations[:, 0] == image])
    val_annotations = np.array(val_annotations, dtype=object)
    csv_path = os.path.join(val_dir, "val_coordinates.csv")
    with open(csv_path, "w", newline="") as f:
        csv.writer(f).writerows(val_annotations)


def train_val_split2(
    traindir: str, images: List[str], image_format: str = "npz", split: float = 0.8
) -> None:
    """After a train/test split has occured, an other train/val split is required inside the
    train subfolder. This could be done on the fly during training but hard separating the files
    seems cleaner and force a constant validation set regardless of the training config.

    Args:
        traindir (str): The train subfolder.
        image_format (str, optional): Used to detect images inside the given folder.
                                      Defaults to 'npz'.
        split (float, optional): Proportion of images to keep for training. Defaults to 0.8.
    """
    parent_dir = os.path.dirname(os.path.abspath(traindir))
    val_dir = os.path.join(parent_dir, "val")
    os.makedirs(val_dir, exist_ok=True)
    for image in images:
        shutil.copyfile(os.path.join(traindir, image), os.path.join(val_dir, image))
    annotations = load_annotations(os.path.join(traindir, "train_coordinates.csv"))
    train_length = int(split * len(annotations))
    train_annotations = annotations[:train_length]
    val_annotations = annotations[train_length:]
    train_annotations = np.array(train_annotations, dtype=object)
    csv_path = os.path.join(traindir, "train_coordinates.csv")
    with open(csv_path, "w", newline="") as f:
        csv.writer(f).writerows(train_annotations)
    val_annotations = np.array(val_annotations, dtype=object)
    csv_path = os.path.join(val_dir, "val_coordinates.csv")
    with open(csv_path, "w", newline="") as f:
        csv.writer(f).writerows(val_annotations)


def train_val_test_splits(
    rootdir: str,
    images_names: List[str],
    train_test_ratio: float,
    train_val_ratio: float,
    image_format: str = "npz",
) -> None:
    train_images = train_test_split(
        rootdir, images_names, image_format, train_test_ratio
    )
    # train_val_split(os.path.join(rootdir, 'train'), train_images, image_format, train_val_ratio)
    train_val_split2(
        os.path.join(rootdir, "train"), train_images, image_format, train_val_ratio
    )


# +------------------------------------------------------------------------------------------+ #
# |                                       EXTERNAL CALL                                      | #
# +------------------------------------------------------------------------------------------+ #


def prepare(
    rootdir: str,
    extension: str,
    pointcloud_path: str = None,
    size: int = None,
    make_u_masks: bool = False,
    train_test_split: float = None,
    train_val_split: float = None,
    crop_output_dir: str = None,
    crop_size: Tuple[int, ...] = None,
    margin: Tuple[int, ...] = 0,
    pos_ratio: float = 1,
    positive_only: bool = False,
    downscale: float = 1.0,
) -> None:
    """Being given a path to a directory, prepare the image for training by doing any
    combinations of the following operations:
    1. generate synthetic data
    2. make u masks
    3. train/val/test splits
    4. make crops

    Args:
        rootdir (str): Directory which (will) contain images and annotations.
        extension (str, optional): Images format.
        size (int, optional): Number of images to generate.
        make_u_masks (bool): If True, will generate one U mask per image in a subfolder
                             rootdir/U_masks/. Defaults to False.
        train_test_split (float, optional): First, the (train+val)/test split will occur.
                                            This determines the train/test number of images ratio.
                                            Defaults to None.
        train_val_split (float, optional): After the (train+val)/test split, the (train+val)
                                           images will be splitted again. Defaults to None.

        crop_size (Tuple[int], optional): Defaults to (64, 128, 128).
        max_particle_dim (int, optional): Defaults to 50.
        margin (Tuple[int], optional): Crop will be randomly shifted so that they are not always
                                       centered on a particle. The margin controls how close to a
                                       border the particle can be after the shift.
                                       Defaults to (0, 0, 0).
        pos_ratio (float, optional): Positive/Negative crop ratio. Defaults to 1..
        positive_only (bool, optional): Crop only positive. Defaults to False.
        downscale (float, optional): Downscale factor. Default to 1 (no downscale)
    """
    summary(locals(), "DATA PREPARATION")
    # STEP 1: Data generation
    print("\nPreparing data for training ...\n")
    if size is not None:
        print("| Generating dataset ...")
        try:
            from .generate import generate_data
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                str(e)
                + "\n\nCannot generate synthetic data, you may need the pyfigtree module for that. \
Try `pip install spfluo[figtree]` or `pip install pyfigtree`."
            )
        generate_data(size, rootdir, pointcloud_path, extension)
    # STEP 2: Train/Val/Test splits
    if train_test_split is not None and train_val_split is not None:
        images_names = list(
            filter(lambda x: x.endswith(extension), os.listdir(rootdir))
        )
        # images_names = [str(i)+'.'+extension for i in range(size)]
        print("| Splitting data into train, val, and test subsets ...")
        train_val_test_splits(
            rootdir, images_names, train_test_split, train_val_split, extension
        )
    # STEP 3: Make U masks
    if make_u_masks:
        print("| Creating U masks ...")
        make_U_masks(os.path.join(rootdir, "train"), extension)
        make_U_masks(os.path.join(rootdir, "val"), extension)
    # STEP 4: Make crops
    if crop_output_dir is not None:
        print("| Generating crops inside images ...")
        crop_kwargs = {
            "crop_size": tuple(np.rint(np.array(crop_size) / downscale).astype(int)),
            "margin": margin,
            "pos_ratio": pos_ratio,
            "positive_only": positive_only,
            "extension": extension,
        }
        for d in ["train", "val", "test"]:
            crop_kwargs["csv_name"] = d + "_coordinates"
            subdir = os.path.join(rootdir, d)
            if os.path.exists(subdir):
                crop(subdir, crop_output_dir, **crop_kwargs)
            else:
                print(f"{subdir} not found")



import numpy as np
import pytest
from scipy.ndimage import affine_transform, convolve

from spfluo import data
from spfluo.utils.transform import get_transform_matrix
from spfluo.utils.volume import are_volumes_aligned


def test_generation():
    d = data.generated_anisotropic()
    D, H, W = d["gt"].shape
    dtype = d["gt"].dtype
    N, x = d["poses"].shape
    assert x == 6
    assert d["volumes"].shape == (N, D, H, W)
    assert d["volumes"].dtype == dtype
    assert d["psf"].dtype == dtype


@pytest.mark.slow
@pytest.mark.parametrize(
    "generated_data", [data.generated_anisotropic(), data.generated_isotropic()]
)
def test_poses(generated_data):
    gt = generated_data["gt"]
    for particle, pose in zip(generated_data["volumes"], generated_data["poses"]):
        # H go from gt to particle (which is transformed)
        H = get_transform_matrix(particle.shape, pose[:3], pose[3:], degrees=True)

        # invert this because scipy's affine_transform works backward
        invH = np.linalg.inv(H)
        transformed_gt = affine_transform(gt, invH, order=1)

        # Apply the data model
        transformed_gt_blurred = convolve(
            transformed_gt, generated_data["psf"], mode="constant", cval=0.0
        )

        # Is the transformed groundtruth aligned with particle ?
        assert are_volumes_aligned(particle, transformed_gt_blurred)
